# HEADER
#                    arg/GUI/View/argDataWidget.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
# HEADER


import os

from PySide2.QtCore import Qt, Slot
from PySide2.QtGui import QIcon
from PySide2.QtWidgets import QApplication, QFileDialog, QGridLayout, QGroupBox, QHBoxLayout, QVBoxLayout, QLabel, \
    QLineEdit, QPushButton, QSpacerItem, QSizePolicy, QTableWidget, QTableWidgetItem, QToolButton, QWidget

from arg.GUI.View.argGuiTools import cleanParametersTabwidget


class argDataWidget(QWidget):
    """A widget class to cover 'Data' tab
    """

    def __init__(self):
        super().__init__()

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        scriptDirectory = os.path.dirname(os.path.realpath(__file__))
        icon = QIcon("{}/{}".format(scriptDirectory, "../Graphics/open.png"))

        # Data folder
        self.dataFolderLabel = QLabel(settings.dataFolderValue(settings.label))
        self.dataFolderLineEdit = QLineEdit()
        self.dataFolderOpenButton = QToolButton()
        self.dataFolderOpenButton.setIcon(icon)
        self.dataFolderHLayout = QHBoxLayout()
        self.dataFolderHLayout.addWidget(self.dataFolderLineEdit)
        self.dataFolderHLayout.addWidget(self.dataFolderOpenButton)

        # CAD Parameters
        self.cadParametersGroupBox = QGroupBox(settings.cadParametersSectionKey(), self)
        self.cadParameterLayout = QGridLayout()
        self.cadParametersGroupBox.setLayout(self.cadParameterLayout)
        self.geometryRootLabel = QLabel(settings.geometryRootValue(settings.label))
        self.geometryRootLineEdit = QLineEdit()
        self.geometryRootOpenButton = QToolButton()
        self.geometryRootOpenButton.setIcon(icon)
        self.geometryRootHLayout = QHBoxLayout()
        self.geometryRootHLayout.addWidget(self.geometryRootLineEdit)
        self.geometryRootHLayout.addWidget(self.geometryRootOpenButton)
        self.geometryReportedCADMetadaLabel = QLabel(settings.reportedCadMetaDataValue(settings.label))
        self.geometryReportedCADMetadaLineEdit = QLineEdit()
        self.cadParameterLayout.addWidget(self.geometryRootLabel, 0, 0)
        self.cadParameterLayout.addLayout(self.geometryRootHLayout, 0, 2)
        self.cadParameterLayout.addWidget(self.geometryReportedCADMetadaLabel, 1, 0)
        self.cadParameterLayout.addWidget(self.geometryReportedCADMetadaLineEdit, 1, 2)

        # FEM Parameters
        self.femParametersGroupBox = QGroupBox(settings.femParametersSectionKey(), self)
        self.femParameterLayout = QGridLayout()
        self.femParametersGroupBox.setLayout(self.femParameterLayout)
        self.inputDeckLabel = QLabel(settings.inputDeckValue(settings.label))
        self.inputDeckLineEdit = QLineEdit()
        self.inputDeckOpenButton = QToolButton()
        self.inputDeckOpenButton.setIcon(icon)
        self.inputDeckOpenButton.clicked.connect(self.onInputDeckButtonTriggered)
        self.inputDeckHLayout = QHBoxLayout()
        self.inputDeckHLayout.addWidget(self.inputDeckLineEdit)
        self.inputDeckHLayout.addWidget(self.inputDeckOpenButton)
        self.ignoredBlocksLabel = QLabel(settings.ignoredBlocksValue(settings.label))
        self.ignoredBlocksLineEdit = QLineEdit()
        self.femParameterLayout.addWidget(self.inputDeckLabel, 0, 0)
        self.femParameterLayout.addLayout(self.inputDeckHLayout, 0, 2)
        self.femParameterLayout.addWidget(self.ignoredBlocksLabel, 3, 0)
        self.femParameterLayout.addWidget(self.ignoredBlocksLineEdit, 3, 2)

        # Mappings
        self.mappingsGroupBox = QGroupBox(settings.mappingsSectionKey(), self)
        self.mappingsLayout = QVBoxLayout()
        self.mappingsGroupBox.setLayout(self.mappingsLayout)
        self.CADToFEMLabel = QLabel(settings.cadToFemValue(settings.label))

        # Tab widget
        self.CADToFEMTableWidget = QTableWidget(self)
        self.CADToFEMTableWidget.setColumnCount(3)
        self.CADToFEMTableWidget.setHorizontalHeaderLabels(settings.cadToFemHeaderLabelsList())
        self.CADToFEMTableWidget.showGrid()
        self.CADToFEMTableWidget.resizeColumnsToContents()

        # Add CAD to FEM push button
        self.addCADToFEMPushButton = QPushButton(settings.cadToFemAddKey())

        # Remove CAD to FEM push button
        self.removeCADToFEMPushButton = QPushButton(settings.cadToFemRemoveKey())

        # Button layout
        self.CADToFEMButtonsLayout = QHBoxLayout()
        self.CADToFEMButtonsLayout.addWidget(self.addCADToFEMPushButton)
        self.CADToFEMButtonsLayout.addWidget(self.removeCADToFEMPushButton)
        self.CADToFEMButtonsLayout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Ignored))

        self.FEMToCADLabel = QLabel(settings.femToCadValue(settings.label))

        # Tab widget
        self.FEMToCADTableWidget = QTableWidget(self)
        self.FEMToCADTableWidget.setColumnCount(3)
        self.FEMToCADTableWidget.setHorizontalHeaderLabels(settings.femToCadHeaderLabelsList())
        self.FEMToCADTableWidget.showGrid()
        self.FEMToCADTableWidget.resizeColumnsToContents()

        # Add CAD to FEM push button
        self.addFEMToCADPushButton = QPushButton(settings.femToCadAddKey())

        # Remove CAD to FEM push button
        self.removeFEMToCADPushButton = QPushButton(settings.femToCadRemoveKey())

        # Button layout
        self.FEMToCADButtonsLayout = QHBoxLayout()
        self.FEMToCADButtonsLayout.addWidget(self.addFEMToCADPushButton)
        self.FEMToCADButtonsLayout.addWidget(self.removeFEMToCADPushButton)
        self.FEMToCADButtonsLayout.addItem(QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Ignored))

        # Set Mappings Layout
        self.mappingsLayout.addWidget(self.CADToFEMLabel)
        self.mappingsLayout.addWidget(self.CADToFEMTableWidget)
        self.mappingsLayout.addLayout(self.CADToFEMButtonsLayout)
        self.mappingsLayout.addWidget(self.FEMToCADLabel)
        self.mappingsLayout.addWidget(self.FEMToCADTableWidget)
        self.mappingsLayout.addLayout(self.FEMToCADButtonsLayout)

        # Instantiate 'Data' tab layout and fill with widgets created above
        self.dataLayout = QGridLayout()
        self.dataLayout.setMargin(5)
        self.dataLayout.addWidget(self.dataFolderLabel, 1, 0)
        self.dataLayout.addLayout(self.dataFolderHLayout, 1, 2)
        self.dataLayout.setColumnMinimumWidth(1, 5)

        self.dataMainLayout = QVBoxLayout(self)
        self.dataMainLayout.setMargin(5)
        self.dataMainLayout.addLayout(self.dataLayout)
        self.dataMainLayout.addWidget(self.cadParametersGroupBox)
        self.dataMainLayout.addWidget(self.femParametersGroupBox)
        self.dataMainLayout.addWidget(self.mappingsGroupBox)

        self.dataMainLayout.addItem(QSpacerItem(0, 0, QSizePolicy.Ignored, QSizePolicy.Expanding))
        self.setLayout(self.dataMainLayout)

        # Logical connections
        self.dataFolderOpenButton.clicked.connect(self.onDataFolderButtonTriggered)
        self.geometryRootOpenButton.clicked.connect(self.onGeometryRootButtonTriggered)
        self.addCADToFEMPushButton.clicked.connect(self.onAddCADToFEM)
        self.removeCADToFEMPushButton.clicked.connect(self.onRemoveCADToFEM)
        self.addFEMToCADPushButton.clicked.connect(self.onAddFEMToCAD)
        self.removeFEMToCADPushButton.clicked.connect(self.onRemoveFEMToCAD)

        self.dataFolderLineEdit.editingFinished.connect(self.onDataFolderLineEditEditingFinished)
        self.geometryRootLineEdit.editingFinished.connect(self.onGeometryRootLineEditEditingFinished)
        self.inputDeckLineEdit.editingFinished.connect(self.onInputDeckEditEditingFinished)

    @staticmethod
    def fillTableWidget(bijectiveMappingsKey, tableWidget, data):
        """ Fill table widget with mapping
        """

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        # Retrieve mappings content from provided data
        bijectiveMappings = data.get(settings.bijectiveMappingsKey())
        if bijectiveMappings:

            # Retrieve value
            bijectiveMappingsValue = bijectiveMappings.get(bijectiveMappingsKey)
            if bijectiveMappingsValue:

                # Retrieve corresponding elements
                bijectiveMappingsElements = bijectiveMappingsValue.get(settings.elementsKey())
                if bijectiveMappingsElements:

                    # Retrieve number of elements to add as many rows
                    bijectiveMappingsElementsCounts = len(bijectiveMappingsElements)
                    tableWidget.setRowCount(bijectiveMappingsElementsCounts)
                    i = 0

                    # Display element and value for each element
                    for elementKey, elementValue in bijectiveMappingsElements.items():
                        tableWidget.setItem(i, 0, QTableWidgetItem(elementKey))
                        tableWidget.setItem(i, 1, QTableWidgetItem(
                            settings.keySeparatorDefault().join(elementValue)))

                        i = i + 1

                # Retrieve corresponding factors
                bijectiveMappingsFactors = bijectiveMappingsValue.get(settings.factorsKey())
                if bijectiveMappingsFactors:

                    # Display factor and value for each element
                    for factorKey, factorValue in bijectiveMappingsFactors.items():
                        for item in tableWidget.findItems(factorKey, Qt.MatchExactly):
                            if item.column() == 0:
                                tableWidget.setItem(item.row(), 2,
                                                    QTableWidgetItem("{}".format(factorValue)))

    def fillParameters(self, data):
        """Set all 'Data' tab widgets to specified values
        """

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        # Clean all widgets before filling them with appropriate values
        self.cleanParameters()

        # Data folder
        if data.get(settings.dataFolderKey()):
            self.dataFolderLineEdit.setText(data.get(settings.dataFolderKey()))

        # CAD Parameters
        if data.get(settings.geometryRootKey()):
            self.geometryRootLineEdit.setText(data.get(settings.geometryRootKey()))
        if data.get(settings.reportedCadMetaDataKey()):
            cadMetaDataValue = data.get(settings.reportedCadMetaDataKey())
            cadMetaDataValueAsString = settings.keySeparatorDefault().join(cadMetaDataValue)
            self.geometryReportedCADMetadaLineEdit.setText(cadMetaDataValueAsString)

        if data.get(settings.inputDeckKey()):
            self.inputDeckLineEdit.setText(data.get(settings.inputDeckKey()))

        if data.get(settings.ignoredBlocksKey()):
            ignoredBlocksValue = data.get(settings.ignoredBlocksKey())
            ignoredBlocksAsString = settings.keySeparatorDefault().join(ignoredBlocksValue)
            self.ignoredBlocksLineEdit.setText(ignoredBlocksAsString)

        # Mappings
        self.fillTableWidget(settings.cadToFemKey(), self.CADToFEMTableWidget, data)
        self.fillTableWidget(settings.femToCadKey(), self.FEMToCADTableWidget, data)

    @staticmethod
    def constructParametersFromTabWidget(tabWidget, subParameter):
        """Complete provided tabWidget with subParameter content
        """

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        elements = {}
        factors = {}
        for i in range(0, tabWidget.rowCount()):
            currentElementItem = tabWidget.item(i, 0)
            currentElementBijectiveItem = tabWidget.item(i, 1)
            currentFactorItem = tabWidget.item(i, 2)

            # If col(0) item exists
            if currentElementItem:
                currentElementStr = currentElementItem.text()

                # If col(1) item item exists
                if currentElementBijectiveItem:
                    currentElementBijectiveCollection = \
                        currentElementBijectiveItem.text().split(settings.keySeparatorDefault())

                    # If col(0) and col(1) item text isn't empty
                    if currentElementStr and (len(currentElementBijectiveCollection) > 0):
                        elements[currentElementStr] = currentElementBijectiveCollection

                # If col(2) item exists
                if currentFactorItem:
                    currentFactorStr = currentFactorItem.text()
                    if currentElementStr and currentFactorStr:
                        factors[currentElementStr] = currentFactorStr

        if len(elements) > 0:
            subParameter[settings.elementsKey()] = elements
        if len(factors) > 0:
            subParameter[settings.factorsKey()] = factors

    def constructParameters(self):
        """Instantiate all 'Data' tab widgets and set to default
        """

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        # Initialize parameters dict
        parameters = {
            settings.dataFolderKey(): self.dataFolderLineEdit.text() if self.dataFolderLineEdit.text()
            else settings.dataFolderDefault(),
            settings.geometryRootKey(): self.geometryRootLineEdit.text()}

        # Data folder

        # CAD Parameters
        cadMetaDataList = self.convertStringAsList(self.geometryReportedCADMetadaLineEdit.text())
        if len(cadMetaDataList) > 0:
            parameters[settings.reportedCadMetaDataKey()] = cadMetaDataList

        parameters[settings.inputDeckKey()] = self.inputDeckLineEdit.text()
        ignoredBlocksList = self.convertStringAsList(self.ignoredBlocksLineEdit.text())
        if len(ignoredBlocksList) > 0:
            parameters[settings.ignoredBlocksKey()] = ignoredBlocksList

        # Mappings
        mappingsParameters = {}
        CADToFEMParameters = {}
        FEMToCADParameters = {}
        self.constructParametersFromTabWidget(self.CADToFEMTableWidget, CADToFEMParameters)
        self.constructParametersFromTabWidget(self.FEMToCADTableWidget, FEMToCADParameters)
        if len(CADToFEMParameters) > 0:
            mappingsParameters[settings.cadToFemKey()] = CADToFEMParameters
        if len(FEMToCADParameters) > 0:
            mappingsParameters[settings.femToCadKey()] = FEMToCADParameters
        if len(mappingsParameters) > 0:
            parameters[settings.bijectiveMappingsKey()] = mappingsParameters

        # Return completed parameters dict
        return parameters

    def cleanParameters(self):
        """Reset all 'Data' tab widgets to default
        """

        self.dataFolderLineEdit.setText("")

        # CAD Parameters
        self.geometryRootLineEdit.setText("")
        self.geometryReportedCADMetadaLineEdit.setText("")

        # FEM Parameters
        self.inputDeckLineEdit.setText("")
        self.ignoredBlocksLineEdit.setText("")

        # Mappings
        cleanParametersTabwidget(self.FEMToCADTableWidget)
        cleanParametersTabwidget(self.CADToFEMTableWidget)

    @staticmethod
    def convertStringAsList(text):
        """Convert a string provided as a list. The string is splitted using
        the defaut key separator provied by the settings controller.
        """

        # Retrieve settings controller
        settings = QApplication.instance().settingsController

        collection = []
        if len(text) > 0:
            collection = text.split(settings.keySeparatorDefault())

        return collection

    @Slot()
    def onAddCADToFEM(self):
        """ Add an CADToFEM to the tabwidget.
        """
        nbCADToFEM = self.CADToFEMTableWidget.rowCount()
        self.CADToFEMTableWidget.setRowCount(nbCADToFEM + 1)

        self.CADToFEMTableWidget.setItem(nbCADToFEM, 0, QTableWidgetItem(""))
        self.CADToFEMTableWidget.setItem(nbCADToFEM, 1, QTableWidgetItem(""))
        self.CADToFEMTableWidget.setItem(nbCADToFEM, 2, QTableWidgetItem(""))

        self.CADToFEMTableWidget.resizeColumnsToContents()

    @Slot()
    def onRemoveCADToFEM(self):
        """ Remove an CADToFEM to the tabwidget.
        """
        self.onRemoveTableWidgetSelectedItems(self.CADToFEMTableWidget)

    @Slot()
    def onAddFEMToCAD(self):
        """ Add an FEMToCAD to the tabwidget.
        """
        nbFEMToFEM = self.FEMToCADTableWidget.rowCount()
        self.FEMToCADTableWidget.setRowCount(nbFEMToFEM + 1)

        self.FEMToCADTableWidget.setItem(nbFEMToFEM, 0, QTableWidgetItem(""))
        self.FEMToCADTableWidget.setItem(nbFEMToFEM, 1, QTableWidgetItem(""))
        self.FEMToCADTableWidget.setItem(nbFEMToFEM, 2, QTableWidgetItem(""))

        self.FEMToCADTableWidget.resizeColumnsToContents()

    @Slot()
    def onRemoveFEMToCAD(self):
        """ Remove an FEMToCAD to the tabwidget.
        """
        self.onRemoveTableWidgetSelectedItems(self.FEMToCADTableWidget)

    @staticmethod
    def onFillLineEdit(lineEdit, fileMode):
        openFileDialog = QFileDialog()
        openFileDialog.setAcceptMode(QFileDialog.AcceptOpen)
        openFileDialog.setFileMode(fileMode)
        openFileDialog.setViewMode(QFileDialog.Detail)
        if openFileDialog.exec_():
            fileNames = openFileDialog.selectedFiles()
            if len(fileNames) == 1:
                lineEdit.setText(fileNames[0])

    @staticmethod
    def onRemoveTableWidgetSelectedItems(tabWidget):
        """Remove selected row from table
        """

        rows = []
        for selectedItem in tabWidget.selectedItems():
            rows.append(selectedItem.row())

        # Remove duplicated
        rows = list(set(rows))

        # sort from higher to lower
        rows.sort(reverse=True)

        # remove sorted rows
        for row in rows:
            tabWidget.removeRow(row)

            tabWidget.resizeColumnsToContents()

    def onDataFolderButtonTriggered(self):
        self.onFillLineEdit(self.dataFolderLineEdit, QFileDialog.DirectoryOnly)
        self.onDataFolderLineEditEditingFinished()

    def onGeometryRootButtonTriggered(self):
        self.onFillLineEdit(self.geometryRootLineEdit, QFileDialog.DirectoryOnly)
        self.onGeometryRootLineEditEditingFinished()

    def onInputDeckButtonTriggered(self):
        self.onFillLineEdit(self.inputDeckLineEdit, QFileDialog.AnyFile)
        self.onInputDeckEditEditingFinished()

    def onDataFolderLineEditEditingFinished(self):
        QApplication.instance().checkLineEdit(self.dataFolderLineEdit, True)

    def onGeometryRootLineEditEditingFinished(self):
        QApplication.instance().checkLineEdit(self.geometryRootLineEdit, True)

    def onInputDeckEditEditingFinished(self):
        QApplication.instance().checkLineEdit(self.inputDeckLineEdit, False)



#HEADER
#                     arg/GUI/Logic/argSettingsController.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import os
import sys

from PySide2.QtCore import QObject

from arg.GUI.Logic.argUserSettingsReader import argUserSettingsReader
from arg.GUI.Logic.argUserSettingsWriter import argUserSettingsWriter


class argSettingsController(QObject):
    """A controller class to handle settings
    """

    def __init__(self):
        super().__init__()
        self.label = 0
        self.mandatory = 1
        self.commentary = 2
        self.settings = {}
        self.confSettingsFileName = "ARG-GUI-config.yml"
        self.confSettingsFilePath = 'Logic/'
        self.scriptDirectory = os.path.dirname(os.path.realpath(__file__))
        self.scriptDirectoryConfigFilePath = os.path.join(self.scriptDirectory, self.confSettingsFileName)
        self.homePath = os.environ["APPDATA"] if "APPDATA" in os.environ else os.environ["HOME"]
        self.userHomeConfigFilePath = os.path.join(os.path.join(self.homePath, "ARG"), self.confSettingsFileName)
        self.userSettingsReader = argUserSettingsReader(self.scriptDirectoryConfigFilePath, self.userHomeConfigFilePath)
        self.userSettingsWriter = argUserSettingsWriter()
        self.argPythonExePath = None
        self.argPythonSitePackagePath = None
        self.argScriptPath = None
        self.argLatexProcessorPath = None

        # Current file opened
        self.currentParameterFile = ""

        # Current file run
        self.currentParameterFileRun = ""

    def initialize(self):

        # Initialize user settings from config file
        missingSettings = self.initializeUserSettings()

        # Report information
        self.settings[self.backendTypeKey()] = ["Backend", True]
        self.settings[self.reportTypeKey()] = ["Report Type", True]
        self.settings[self.classificationKey()] = ["Classification", True]
        self.settings[self.fileNameKey()] = ["File Name", False]
        self.settings[self.mutablesFileKey()] = ["Mutables File", False]
        self.settings[self.structureFileKey()] = ["Structure File", False]
        self.settings[self.structureEndFileKey()] = ["Analyst Authored Results Section", False]
        self.settings[self.artifactFileKey()] = ["Artifact File", False]
        self.settings[self.outputFolderKey()] = ["Output Folder", True]
        self.settings[self.verbosityKey()] = ["Verbosity", False]

        # General Options
        self.settings[self.titleKey()] = ["Title", False]
        self.settings[self.numberKey()] = ["Number", False]
        self.settings[self.issueKey()] = ["Issue", False]
        self.settings[self.versionsKey()] = ["Versions", False]
        self.settings[self.authorsKey()] = ["Authors", False]
        self.settings[self.organizationsKey()] = ["Organizations", False]
        self.settings[self.locationKey()] = ["Location", False]
        self.settings[self.yearKey()] = ["Year", False]
        self.settings[self.monthKey()] = ["Month", False]
        self.settings[self.abstractFileKey()] = ["Abstract File", False]
        self.settings[self.prefaceKey()] = ["Preface", False]
        self.settings[self.thanksKey()] = ["Thanks", False]
        self.settings[self.executiveSummaryKey()] = ["Executive Summary", False]
        self.settings[self.nomenclatureKey()] = ["Nomenclature", False]
        self.settings[self.finalKey()] = ["Final", False]
        self.settings[self.keySeparatorKey()] = ["Key Separator", False]

        # Data
        self.settings[self.dataFolderKey()] = ["Model Directory", False]
        self.settings[self.inputDeckKey()] = ["Input Deck", False]
        self.settings[self.geometryRootKey()] = ["Geometry Root", False]
        self.settings[self.reportedCadMetaDataKey()] = ["Reported CAD Metadata", False]
        self.settings[self.logFileKey()] = ["Log File", False]
        self.settings[self.ignoredBlocksKey()] = ["Ignored Blocks", False]
        self.settings[self.bijectiveMappingsKey()] = ["Bijective Mapping", False]
        self.settings[self.cadToFemKey()] = ["CAD to FEM", False]
        self.settings[self.femToCadKey()] = ["FEM to CAD", False]

        # Return list of missing settings
        return missingSettings

    def initializeUserSettings(self):

        # Read settings
        missingAdmin, missingSettings = self.userSettingsReader.read()

        # Exit early if ADMIN level config file is missing
        if missingAdmin:
            print("** ERROR: No complete config file at ADMIN level \'{}\'. Exiting. ".format(
                self.scriptDirectoryConfigFilePath))
            sys.exit(1)
        elif missingSettings:
            return missingSettings

        # Python executable
        self.argPythonExePath = self.userSettingsReader.getPythonExecutable().replace('\\', '/')
        print("\t{} = {}".format(self.getArgPythonExePathUserSettingsKey(), self.argPythonExePath))

        # Python packages
        self.argPythonSitePackagePath = self.userSettingsReader.getPythonSitePackage().replace('\\', '/')
        print("\t{} = {}".format(self.getArgPythonSitePackagePathUserSettingsKey(), self.argPythonSitePackagePath))

        # ARG.py script
        self.argScriptPath = self.userSettingsReader.getArgScript().replace('\\', '/')
        print("\t{} = {}".format(self.getArgScriptPathUserSettingsKey(), self.argScriptPath))

        # LaTeX processor
        self.argLatexProcessorPath = self.userSettingsReader.getLatexProcessor().replace('\\', '/')
        print("\t{} = {}".format(self.getArgLatexProcessorPathUserSettingsKey(), self.argLatexProcessorPath))

    def saveSettings(self, data):

        # Write current set of settings to config file at script directory location
        if self.userSettingsWriter.write(self.userHomeConfigFilePath, data):

            # Save current set of settings as RunTime paths
            self.setArgPythonExePath(data[self.getArgPythonExePathUserSettingsKey()])
            self.setArgPythonSitePackagePath(data[self.getArgPythonSitePackagePathUserSettingsKey()])
            self.setArgScriptPath(data[self.getArgScriptPathUserSettingsKey()])
            self.setArgLatexProcessorPath(data[self.getArgLatexProcessorPathUserSettingsKey()])

        else:
            print("*  WARNING: Current set of settings could not be saved:\n\t{}".format(data))

    @staticmethod
    def getCompanyName():
        return "ARG"

    @staticmethod
    def getToolName():
        return "ARG-GUI"

    @staticmethod
    def getRecentMenuSettings():
        return "Recent Menu"

    @staticmethod
    def getSaveBeforeRunSettings():
        return "SaveBeforeRunEnabled"

    def getConfigArgPythonExePath(self):
        return self.userSettingsReader.getPythonExecutable()

    def readConfigArgPythonExePath(self):
        return self.userSettingsReader.readPythonExecutable(self.userSettingsReader.usrLvlPath)

    def getArgPythonExePath(self):
        return self.argPythonExePath

    @staticmethod
    def getArgPythonExePathLabel():
        return "Python executable"

    @staticmethod
    def getArgPythonExePathKey():
        return "argPythonExePath"

    @staticmethod
    def getArgPythonExePathUserSettingsKey():
        return "python_executable"

    @staticmethod
    def getArgPythonExePathToolTip():
        return "Fill the line edit with the complete path of the Python executable to be used to run ARG."

    def setArgPythonExePath(self, pythonExePath):
        self.argPythonExePath = pythonExePath

    def getConfigArgPythonSitePackagePath(self):
        return self.userSettingsReader.getPythonSitePackage()

    def readConfigArgPythonSitePackagePath(self):
        return self.userSettingsReader.readPythonSitePackage(self.userSettingsReader.usrLvlPath)

    def getArgPythonSitePackagePath(self):
        return self.argPythonSitePackagePath

    @staticmethod
    def getArgPythonSitePackagePathLabel():
        return "Python site-packages folder"

    @staticmethod
    def getArgPythonSitePackagePathKey():
        return "argPythonSitePackagePath"

    @staticmethod
    def getArgPythonSitePackagePathUserSettingsKey():
        return "python_site_package"

    @staticmethod
    def getArgPythonSitePackagePathToolTip():
        return "Fill the line edit with the complete path to the site-packages directory required by Python. <br><br>" \
               "(e.g: '<code>.../Python.framework/Versions/3.7/lib/python3.7/site-packages</code>')"

    def setArgPythonSitePackagePath(self, pythonSitePackagePath):
        self.argPythonSitePackagePath = pythonSitePackagePath

    def getConfigArgScriptPath(self):
        return self.userSettingsReader.getArgScript()

    def readConfigArgScriptPath(self):
        return self.userSettingsReader.readArgScript(self.userSettingsReader.usrLvlPath)

    @staticmethod
    def getArgScriptPathLabel():
        return "ARG.py script"

    @staticmethod
    def getArgScriptPathKey():
        return "argScriptPath"

    @staticmethod
    def getArgScriptPathUserSettingsKey():
        return "arg_script"

    @staticmethod
    def getArgScriptPathToolTip():
        return "Fill the line edit with the complete path to '<code>ARG.py</code>' script."

    def getArgScriptPath(self):
        return self.argScriptPath

    def setArgScriptPath(self, scriptPath):
        self.argScriptPath = scriptPath

    def getConfigArgLatexProcessorPath(self):
        return self.userSettingsReader.getLatexProcessor()

    def readConfigArgLatexProcessorPath(self):
        return self.userSettingsReader.readLatexProcessor(self.userSettingsReader.usrLvlPath)

    def getArgLatexProcessorPath(self):
        return self.argLatexProcessorPath

    @staticmethod
    def getArgLatexProcessorPathLabel():
        return "Latexmk processor"

    @staticmethod
    def getArgLatexProcessorPathKey():
        return "argLatexProcessorPath"

    @staticmethod
    def getArgLatexProcessorPathUserSettingsKey():
        return "latex_processor"

    @staticmethod
    def getArgLatexProcessorPathToolTip():
        return "Fill the line edit with the complete path to Latexmk processor executable.<br><br>" \
               "(e.g: '<code>.../bin/latexmk</code>')<br><br><br><i><u>Note</u>: The <code>Latexmk</code> processor " \
               "is not part of ARG, the ability to call it from ARG is provided as a convenience.</i>"

    def setArgLatexProcessorPath(self, latexProcessorPath):
        self.argLatexProcessorPath = latexProcessorPath

    def getCurrentParameterFile(self):
        return self.currentParameterFile

    def setCurrentParameterFile(self, filePath):
        self.currentParameterFile = filePath

    def getCurrentParameterFileRun(self):
        return self.currentParameterFileRun

    def setCurrentParameterFileRun(self, filePath):
        self.currentParameterFileRun = filePath

    def settingsValue(self, keyFunction, keyValue):
        app = 'ARG-GUI'
        if keyValue is None:
            return self.settings.get(keyFunction())
        elif keyValue < 3:
            return self.settings.get(keyFunction())[keyValue]
        else:
            print("[{}] Could not find value for parameter \'{}\'.".format(app, keyFunction))
            return ""

    # ARG execution option
    @staticmethod
    def eOption():
        return "-e"

    @staticmethod
    def gOption():
        return "-g"

    # Report information
    @staticmethod
    def backendTypeKey():
        return "BackendType"

    @staticmethod
    def backendTypeParameterFileKey():
        return "backend_type"

    def backendTypeValue(self, keyValue=None):
        return self.settingsValue(self.backendTypeKey, keyValue)

    @staticmethod
    def backendTypeKeyWord():
        return "Word"

    @staticmethod
    def backendTypeKeyLatex():
        return "LaTeX"

    @staticmethod
    def reportTypeKeyReport():
        return "Report"

    @staticmethod
    def reportTypeKey():
        return "ReportType"

    @staticmethod
    def reportTypeParameterFileKey():
        return "report_type"

    def reportTypeValue(self, keyValue=None):
        return self.settingsValue(self.reportTypeKey, keyValue)

    @staticmethod
    def classificationKey():
        return "Classification"

    @staticmethod
    def classificationParameterFileKey():
        return "classification"

    @staticmethod
    def classificationKeyGeneric():
        return "Generic"

    def classificationValue(self, keyValue=None):
        return self.settingsValue(self.classificationKey, keyValue)

    @staticmethod
    def fileNameKey():
        return "File Name"

    @staticmethod
    def fileNameParameterFileKey():
        return "file_name"

    def fileNameValue(self, keyValue=None):
        return self.settingsValue(self.fileNameKey, keyValue)

    @staticmethod
    def mutablesFileKey():
        return "Mutables"

    @staticmethod
    def mutablesFileParameterFileKey():
        return "mutables"

    def mutablesFileValue(self, keyValue=None):
        return self.settingsValue(self.mutablesFileKey, keyValue)

    @staticmethod
    def structureFileKey():
        return "StructureFile"

    @staticmethod
    def structureFileParameterFileKey():
        return "structure"

    def structureFileValue(self, keyValue=None):
        return self.settingsValue(self.structureFileKey, keyValue)

    @staticmethod
    def structureEndFileKey():
        return "StructureEnd"

    @staticmethod
    def structureEndFileParameterFileKey():
        return "structure_end"

    def structureEndFileValue(self, keyValue=None):
        return self.settingsValue(self.structureEndFileKey, keyValue)

    @staticmethod
    def artifactFileKey():
        return "ArtifactFile"

    @staticmethod
    def artifactFileParameterFileKey():
        return "artifact"

    def artifactFileValue(self, keyValue=None):
        return self.settingsValue(self.artifactFileKey, keyValue)

    @staticmethod
    def outputFolderKey():
        return "OutputDir"

    @staticmethod
    def outputFolderParameterFileKey():
        return "output"

    def outputFolderValue(self, keyValue=None):
        return self.settingsValue(self.outputFolderKey, keyValue)

    @staticmethod
    def verbosityKey():
        return "Verbosity"

    @staticmethod
    def verbosityParameterFileKey():
        return "verbosity"

    @staticmethod
    def verbosityKeyTerse():
        return "Terse"

    @staticmethod
    def verbosityKeyTerseAsInt():
        return -1

    @staticmethod
    def verbosityKeyDefault():
        return "Default"

    @staticmethod
    def verbosityKeyDefaultAsInt():
        return 0

    @staticmethod
    def verbosityKeyVerbose():
        return "Verbose"

    @staticmethod
    def verbosityKeyVerboseAsInt():
        return 1

    def verbosityValue(self, keyValue=None):
        return self.settingsValue(self.verbosityKey, keyValue)

    # General Options
    @staticmethod
    def titleKey():
        return "Title"

    def titleValue(self, keyValue=None):
        return self.settingsValue(self.titleKey, keyValue)

    @staticmethod
    def titleParameterFileKey():
        return "title"

    @staticmethod
    def numberKey():
        return "Number"

    def numberValue(self, keyValue=None):
        return self.settingsValue(self.numberKey, keyValue)

    @staticmethod
    def numberParameterFileKey():
        return "number"

    @staticmethod
    def issueKey():
        return "Issue"

    def issueValue(self, keyValue=None):
        return self.settingsValue(self.issueKey, keyValue)

    @staticmethod
    def issueParameterFileKey():
        return "issue"

    @staticmethod
    def versionsKey():
        return "Versions"

    def versionsValue(self, keyValue=None):
        return self.settingsValue(self.versionsKey, keyValue)

    @staticmethod
    def versionParameterFileKey():
        return "version"

    @staticmethod
    def authorsKey():
        return "Authors"

    def authorsValue(self, keyValue=None):
        return self.settingsValue(self.authorsKey, keyValue)

    @staticmethod
    def authorParameterFileKey():
        return "author"

    @staticmethod
    def organizationsKey():
        return "Organizations"

    def organizationsValue(self, keyValue=None):
        return self.settingsValue(self.organizationsKey, keyValue)

    @staticmethod
    def organizationParameterFileKey():
        return "organization"

    @staticmethod
    def locationKey():
        return "Location"

    def locationValue(self, keyValue=None):
        return self.settingsValue(self.locationKey, keyValue)

    @staticmethod
    def locationParameterFileKey():
        return "location"

    @staticmethod
    def yearKey():
        return "Year"

    def yearValue(self, keyValue=None):
        return self.settingsValue(self.yearKey, keyValue)

    @staticmethod
    def yearParameterFileKey():
        return "year"

    @staticmethod
    def monthKey():
        return "Month"

    def monthValue(self, keyValue=None):
        return self.settingsValue(self.monthKey, keyValue)

    @staticmethod
    def monthParameterFileKey():
        return "month"

    @staticmethod
    def abstractFileKey():
        return "AbstractFile"

    def abstractFileValue(self, keyValue=None):
        return self.settingsValue(self.abstractFileKey, keyValue)

    @staticmethod
    def abstractFileParameterFileKey():
        return "abstract"

    @staticmethod
    def prefaceKey():
        return "Preface"

    def prefaceValue(self, keyValue=None):
        return self.settingsValue(self.prefaceKey, keyValue)

    @staticmethod
    def prefaceParameterFileKey():
        return "preface"

    @staticmethod
    def thanksKey():
        return "Thanks"

    def thanksValue(self, keyValue=None):
        return self.settingsValue(self.thanksKey, keyValue)

    @staticmethod
    def thanksParameterFileKey():
        return "thanks"

    @staticmethod
    def executiveSummaryKey():
        return "ExecutiveSummary"

    def executiveSummaryValue(self, keyValue=None):
        return self.settingsValue(self.executiveSummaryKey, keyValue)

    @staticmethod
    def executiveSummaryParameterFileKey():
        return "executive_summary"

    @staticmethod
    def nomenclatureKey():
        return "Nomenclature"

    def nomenclatureValue(self, keyValue=None):
        return self.settingsValue(self.nomenclatureKey, keyValue)

    @staticmethod
    def nomenclatureParameterFileKey():
        return "nomenclature"

    @staticmethod
    def finalKey():
        return "Final"

    def finalValue(self, keyValue=None):
        return self.settingsValue(self.finalKey, keyValue)

    @staticmethod
    def finalParameterFileKey():
        return "final"

    @staticmethod
    def keySeparatorKey():
        return "KeySeparator"

    @staticmethod
    def keySeparatorDefault():
        return ";"

    def keySeparatorValue(self, keyValue=None):
        return self.settingsValue(self.keySeparatorKey, keyValue)

    @staticmethod
    def keySeparatorParameterFileKey():
        return "key_separator"

    # Data
    @staticmethod
    def dataFolderKey():
        return "DataDirectory"

    @staticmethod
    def dataFolderParameterFileKey():
        return "data"

    @staticmethod
    def dataFolderDefault():
        return '.'

    def dataFolderValue(self, keyValue=None):
        return self.settingsValue(self.dataFolderKey, keyValue)

    @staticmethod
    def solutionCasesKey():
        return "SolutionCases"

    @staticmethod
    def solutionCasesParameterFileKey():
        return "solution_cases"

    def solutionCasesValue(self, keyValue=None):
        return self.settingsValue(self.solutionCasesKey, keyValue)

    @staticmethod
    def solutionCasesLabelsList():
        return ["Type", "Ignored Blocks", "Mode minimum value", "Mode maximum value"]

    @staticmethod
    def solutionCasesAddKey():
        return "Add solution case"

    @staticmethod
    def solutionCasesRemoveKey():
        return "Remove solution case"

    @staticmethod
    def solutionCasesTypeKey():
        return "type"

    @staticmethod
    def solutionCasesTypeValues():
        return ["eigen"]

    @staticmethod
    def solutionCasesIgnoredBlocksKey():
        return "ignored_blocks"

    @staticmethod
    def solutionCasesModeKey():
        return "modes"

    @staticmethod
    def cadParametersSectionKey():
        return "CAD Parameters"

    @staticmethod
    def geometryRootKey():
        return "GeometryRoot"

    @staticmethod
    def geometryRootParameterFileKey():
        return "geometry_root"

    def geometryRootValue(self, keyValue=None):
        return self.settingsValue(self.geometryRootKey, keyValue)

    @staticmethod
    def reportedCadMetaDataKey():
        return "ReportedCadMetaData"

    @staticmethod
    def reportedCadMetaDataParameterFileKey():
        return "reported_cad_metadata"

    def reportedCadMetaDataValue(self, keyValue=None):
        return self.settingsValue(self.reportedCadMetaDataKey, keyValue)

    @staticmethod
    def femParametersSectionKey():
        return "FEM Parameters"

    @staticmethod
    def inputDeckKey():
        return "DeckRoot"

    @staticmethod
    def inputDeckParameterFileKey():
        return "input_deck"

    def inputDeckValue(self, keyValue=None):
        return self.settingsValue(self.inputDeckKey, keyValue)

    @staticmethod
    def logFileKey():
        return "LogFile"

    @staticmethod
    def logFileParameterFileKey():
        return "log_file"

    def logFileValue(self, keyValue=None):
        return self.settingsValue(self.logFileKey, keyValue)

    @staticmethod
    def ignoredBlocksKey():
        return "IgnoredBlockKeys"

    @staticmethod
    def ignoredBlocksParameterFileKey():
        return "ignored_blocks"

    def ignoredBlocksValue(self, keyValue=None):
        return self.settingsValue(self.ignoredBlocksKey, keyValue)

    @staticmethod
    def mappingsSectionKey():
        return "Mappings"

    @staticmethod
    def bijectiveMappingsKey():
        return "Mappings"

    def bijectiveMappingsValue(self, keyValue=None):
        return self.settingsValue(self.bijectiveMappingsKey, keyValue)

    @staticmethod
    def bijectiveMappingsParameterFileKey():
        return "mappings"

    @staticmethod
    def cadToFemKey():
        return "CAD_to_FEM"

    def cadToFemValue(self, keyValue=None):
        return self.settingsValue(self.cadToFemKey, keyValue)

    @staticmethod
    def cadToFemAddKey():
        return "Add CAD to FEM"

    @staticmethod
    def cadToFemRemoveKey():
        return "Remove CAD to FEM"

    @staticmethod
    def cadToFemHeaderLabelsList():
        return ["CAD parts", "FEM blocks ", "Factors"]

    @staticmethod
    def femToCadKey():
        return "FEM_to_CAD"

    def femToCadValue(self, keyValue=None):
        return self.settingsValue(self.femToCadKey, keyValue)

    @staticmethod
    def femToCadAddKey():
        return "Add FEM to CAD"

    @staticmethod
    def femToCadRemoveKey():
        return "Remove FEM to CAD"

    @staticmethod
    def femToCadHeaderLabelsList():
        return ["FEM blocks ", "CAD parts", "Factors"]

    @staticmethod
    def elementsKey():
        return "elements"

    @staticmethod
    def factorsKey():
        return "factors"

    # Inserts
    @staticmethod
    def insertionKey():
        return "Insert"

    @staticmethod
    def insertionParameterFileKey():
        return "insert_in"

    def insertionValue(self, keyValue=None):
        return self.settingsValue(self.insertionKey(), keyValue)

    @staticmethod
    def insertionLocationKey():
        return "location"

    @staticmethod
    def insertionTypeTextKey():
        return "string"

    @staticmethod
    def insertionTypeImageKey():
        return "image"

    @staticmethod
    def insertionAddKey():
        return "Add insert"

    @staticmethod
    def insertionRemoveKey():
        return "Remove insert"

    @staticmethod
    def insertHeaderLabelsList():
        return ["Location", "Type", "Text / Image Path"]

    @staticmethod
    def textParameterFileKey():
        return "string"

    def textValue(self, keyValue=None):
        return self.settingsValue(self.textKey, keyValue)

    @staticmethod
    def imageParameterFileKey():
        return "image"

    def imageValue(self, keyValue=None):
        return self.settingsValue(self.imageKey, keyValue)



# HEADER
#                      arg/Applications/Explorator.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
# HEADER

import os
import sys
import time

import yaml

from arg import __version__
from arg.Common.argReportParameters import argReportParameters
from arg.Applications.argExplorator import argExplorator

ARG_VERSION = __version__

# Import ARG modules
if not __package__:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
else:
    sys.path.append("..")

# Load supported types
common_dir = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(common_dir, "../Common/argTypes.yml"),
          'r',
          encoding="utf-8") as t_file:
    Types = yaml.safe_load(t_file)


class exploratorSolution:
    """A class to describe a solution file or partition
    """

    def __init__(self):
        # Default instance variables
        self.Type = None
        self.Name = None
        self.Method = None



def main(types, version=None):
    """ Explorator main method
    """

    # Start stopwatch
    t_start = time.time()

    # Print startup information
    sys_version = sys.version_info
    print("[Explorator] ### Started with Python {}.{}.{}".format(
        sys_version.major,
        sys_version.minor,
        sys_version.micro))

    # Instantiate parameters object from command line arguments
    parameters = argReportParameters("Explorator", version=version, types=types)

    # Parse command line arguments to get parameters file value
    if parameters.parse_line():
        # Execute
        execute(parameters)

    # Print error message if something went wrong
    else:
        print("*  ERROR: cannot parse parameters. Exiting.")

    # End stopwatch
    dt = time.time() - t_start

    # If this point is reached everything went fine
    print("[Explorator] Process completed in {} seconds. ###".format(
        dt))


def execute(parameters):
    """ Explorator execute method
    """

    # Initialize storage for automatically determined values
    mutables = {"version": parameters.Version}

    # Parse parameters file
    print("[Explorator] Parsing parameters file")

    # Retrieve all supported files inside data path tree
    case = argExplorator(parameters)
    case.recursively_search_supported_files(parameters.DataDir, parameters.Verbosity)
    parameters.check_geometry_root(case)
    parameters.check_deck_root(case)
    parameters.check_structure_file()
    backend = parameters.Backend

    # Check whether a standalone mesh was found
    if len(case.ExodusIIFiles):
        # Take first mesh file is some are present
        case.MeshName = case.ExodusIIFiles[0]
        case.MeshType = "ExodusII"

        # Store discovered mesh
        print("[Explorator] Stand-alone {} {} was found.".format(
            case.MeshType,
            case.MeshName))
        case.DiscoveredData[case.MeshType] = " mesh in {}".format(
            backend.generate_text(case.MeshName.replace('\\', '/')))

    # Save computed mutables
    parameters.save_generated_mutables(mutables)

    # Generate document structure from discovered data
    generate_structure_file(parameters, case)


def get_and_comment_property_value(meta_info, info_key):
    """ Convenience method to get property value and comment about it
    """

    # Try to retrieve value corresponding to property key
    info_val = meta_info.get(info_key)
    if info_val:
        print("[Explorator] {} specified by input deck: {}".format(
            info_key.capitalize(),
            ", ".join(info_val) if isinstance(info_val, list) else info_val
        ))
    else:
        print("[Explorator] No {} specified by input deck".format(
            info_key))

    # Return retrieved (possibly None) value
    return info_val


def find_solution_partition(case, sol_stem, sol_method, backend):
    """ Find Exodus II partition corresponding to given solution method
    """

    # Iterate over ExodusII partitions associated with the case
    for f in case.ExodusIIPartitions:
        if f.split('.')[0] == sol_stem:
            # Assign found solution partition
            solution = exploratorSolution()
            solution.Type = "ExodusII"
            solution.Name = f
            solution.Method = sol_method

            # Store discovered solution
            print("[Explorator] Solution {} {} was found".format(
                solution.Type,
                f))
            case.DiscoveredData.setdefault("{} solution".format(solution.Type), []).append(
                " in {}".format(backend.generate_text(solution.Name.replace('\\', '/'))))

            # Store corresponding solution and break out
            case.Solutions.append(solution)
            return


def insert_fragment(yaml_file, frag_k, frag_v, chapter_index, section_index, backend):
    """ Insert text and image fragments into YAML structure file
    """

    # Compute indentation
    indent = ' ' * (4 if section_index else 2)

    # Iterate of text fragments
    if frag_k == "string":
        for txt in frag_v:
            # Insert text
            print("[Explorator] Inserting string `{}` into subdivision {}.{}".format(
                txt if len(txt) < 20
                else "{}[...]{}".format(
                    txt[:10],
                    txt[-10:]),
                chapter_index,
                section_index
            ))
            yaml_file.write("{}- n: paragraph\n".format(indent))
            yaml_file.write("{}  string: {}\n".format(indent, txt))

    # Iterate of image fragments
    elif frag_k == "image":
        for img in frag_v:
            # Insert image
            print("Explorator] Inserting image {} into subdivision {}.{}".format(
                img,
                chapter_index,
                section_index
            ))
            yaml_file.write("{}- n: figure\n".format(indent))
            yaml_file.write("{}  arguments:\n".format(indent))
            yaml_file.write("{}    width: 12cm\n".format(indent))
            yaml_file.write("{}    figure_file: {}\n".format(indent, img))
            img = img.replace(os.path.sep, '/').format(indent)
            yaml_file.write(
                "{}    caption_string: 'File {}'\n".format(indent, backend.generate_text(img)))
            yaml_file.write("{}    label: 'f:{}'\n".format(indent, img))

    # Ignore unknown fragment types
    else:
        print("*  WARNING: ignoring unsupported fragment type: {}".format(
            frag_k))


def insert_introduction_chapter(yaml_file, case, chap_index, backend):
    """ Insert introduction chapter into YAML structure file
    """

    # Retrieve chapter fragments and initialize section index
    fragments = case.Fragments.get(chap_index, {})
    sect_index = 0

    # Create chapter header
    yaml_file.write("# Introduction\n")
    yaml_file.write("- n: chapter_null\n")
    yaml_file.write("  title: Introduction\n")

    # Create introduction text regarding explored data
    yaml_file.write(
        "  string: 'The structure of this report was built by the Explorator component of ARG, which explored the following directory:'\n")
    yaml_file.write("  sections:\n")
    yaml_file.write("  - n: paragraph\n")
    yaml_file.write("    verbatim: '{}'\n".format(case.RealDataDir))

    # Append list of discovered items if any
    if case.DiscoveredData:
        yaml_file.write("  - n: paragraph\n")
        yaml_file.write("    string: 'and discovered the following relevant data:'\n")
        yaml_file.write("  - n: itemize\n")
        yaml_file.write("    items:\n")
        for k, v in case.DiscoveredData.items():
            if type(v) == list:
                for e in v:
                    yaml_file.write("    - string: {}{}\n".format(k, e))
            else:
                yaml_file.write("    - string: {}{}\n".format(k, v))
    else:
        yaml_file.write("  - n: paragraph\n")
        yaml_file.write("    string:  No relevant data was discovered.")

    # Append text fragments if any
    for frag_k, frag_v in fragments.get(sect_index, {}).items():
        insert_fragment(yaml_file, frag_k, frag_v, chap_index, sect_index, backend)


def insert_CAD_chapter(yaml_file, case, verbosity_levels, verbosity, chap_index, backend):
    """ Insert CAD chapter into YAML structure file
    """

    # Retrieve chapter fragments and initialize section index
    fragments = case.Fragments.get(chap_index, {})
    sect_index = 0

    # Create chapter header
    yaml_file.write("# CAD chapter\n")
    yaml_file.write("- n: chapter\n")
    yaml_file.write("  title: Geometry\n")
    yaml_file.write("  string: 'This chapter describes the geometry as specified in the parameters file.'\n")
    yaml_file.write("  sections:\n")
    for frag_k, frag_v in fragments.get(sect_index, {}).items():
        insert_fragment(yaml_file, frag_k, frag_v, chap_index, sect_index, backend)

    # Create CAD/FEM sections when corresponding mappings are available
    if case.Mappings:
        set_elements = {"CAD": "part", "FEM": "block"}
        set_titles = {"CAD": "Geometry", "FEM": "Finite Elements"}
        for src, dst in (("CAD", "FEM"), ("FEM", "CAD")):
            data_type = "{}_to_{}".format(src, dst)
            elements = case.Mappings.get(data_type, {}).get("elements")
            if elements:
                # Create new section
                src_name = set_elements.get(src, "items")
                dst_name = set_elements.get(dst, "items")
                sect_index += 1
                yaml_file.write("  - n: section\n")
                yaml_file.write("    title: {} to {} Elements Mapping\n".format(
                    set_titles.get(src, src),
                    set_titles.get(dst, dst)))
                yaml_file.write("    string: 'This section describes the mapping from {} {}s to {} {}s:'\n".format(
                    src,
                    src_name,
                    dst,
                    dst_name))
                yaml_file.write("    sections:\n")
                yaml_file.write("    - n: aggregate\n")
                yaml_file.write("      name: show_all_mappings\n")
                yaml_file.write("      datatype: %s\n" % data_type)
                yaml_file.write("      dataset: %s\n" % elements)
                yaml_file.write("      src_name: %s\n" % src_name)
                yaml_file.write("      dst_name: %s\n" % dst_name)
                yaml_file.write("      logfile: %s\n" % case.LogFile)
                factors = case.Mappings.get(data_type, {}).get("factors")
                if factors and len(case.GeometryFiles) == len(factors):
                    yaml_file.write("      factors: %s\n" % factors)
                    yaml_file.write("      geometry_root: %s\n" % os.path.dirname(case.GeometryFiles[0]))
                for frag_k, frag_v in fragments.get(sect_index, {}).items():
                    insert_fragment(yaml_file, frag_k, frag_v, chap_index, sect_index, backend)

    # Warn when geometry_root is defined but mappings is missing
    else:
        print("*  WARNING: missing CAD/FEM mapping.")

    # Create new section if STL files are present
    if case.GeometryFiles:
        sect_index += 1
        yaml_file.write("  - n: section\n")
        yaml_file.write("    title: Geometry Files\n")
        yaml_file.write("    string: 'This section provides an overview of the geometry files found in:'\n")
        yaml_file.write("    sections:\n")
        yaml_file.write("    - n: paragraph\n")
        yaml_file.write("      verbatim: {}\n".format(
            os.path.join(case.RealDataDir, os.path.dirname(case.GeometryFiles[0]))))
        # Iterate over STL files
        for stl_file in case.GeometryFiles:
            yaml_file.write("    - n: aggregate\n")
            yaml_file.write("      name: show_mesh_surface\n")
            yaml_file.write("      model: %s\n" % stl_file)
            yaml_file.write("      datatype: vtkSTL\n")
            yaml_file.write("      width: 12cm\n")
            yaml_file.write("      axes: True\n")
            yaml_file.write("      merge: True\n")
        for frag_k, frag_v in fragments.get(sect_index, {}).items():
            insert_fragment(yaml_file, frag_k, frag_v, chap_index, sect_index, backend)

        # Create new section if CAD metadata to report present
        if case.ParametersFiles and case.MetaData:
            sect_index += 1
            yaml_file.write("  - n: section\n")
            yaml_file.write("    title: CAD metadata\n")
            yaml_file.write("    string: 'This section describes the CAD metadata found in:'\n")
            yaml_file.write("    sections:\n")
            yaml_file.write("    - n: paragraph\n")
            yaml_file.write("      verbatim: {}\n".format(
                os.path.join(case.RealDataDir, os.path.dirname(case.GeometryFiles[0]))))
            yaml_file.write("    - n: aggregate\n")
            yaml_file.write("      name: show_CAD_metadata\n")
            yaml_file.write("      datatype: vtkSTL\n")
            yaml_file.write("      metadata: %s\n" % case.MetaData)
            yaml_file.write("      parameters_root: {}\n".format(
                os.path.join(case.RealDataDir, os.path.dirname(case.GeometryFiles[0]))))
            for frag_k, frag_v in fragments.get(sect_index, {}).items():
                insert_fragment(yaml_file, frag_k, frag_v, chap_index, sect_index, backend)


def insert_mesh_chapter(yaml_file, case, verbosity_levels, verbosity, chap_index, backend):
    """ Insert mesh chapter into YAML structure file
    """

    # Retrieve chapter fragments and initialize section index
    fragments = case.Fragments.get(chap_index, {})
    sect_index = 0

    # Create chapter header
    yaml_file.write("# Mesh chapter\n")
    yaml_file.write("- n: chapter\n")
    yaml_file.write("  title: {} Mesh\n".format(
        ' '.join([w[0].upper() + w[1:] for w in case.MeshType.split()])))
    yaml_file.write("  string: 'This chapter describes the {} mesh in:'\n".format(
        case.MeshType))
    yaml_file.write("  sections:\n")
    yaml_file.write("  - n: paragraph\n")
    yaml_file.write("    verbatim: {}\n".format(
        os.path.join(case.RealDataDir, case.MeshName)))
    for frag_k, frag_v in fragments.get(sect_index, {}).items():
        insert_fragment(yaml_file, frag_k, frag_v, chap_index, sect_index, backend)

    # Create new section
    sect_index += 1
    yaml_file.write("  - n: section\n")
    yaml_file.write("    title: Overview\n")
    yaml_file.write(
        "    string: This section provides an overview of the meta-data and global properties of this {} mesh.\n".format(
            case.MeshType))
    yaml_file.write("    sections:\n")
    yaml_file.write("    - n: meta\n")
    yaml_file.write("      datatype: %s\n" % case.MeshType)
    yaml_file.write("      dataset: %s\n" % case.MeshName)
    yaml_file.write("    - n: aggregate\n")
    yaml_file.write("      name: show_mesh_surface\n")
    yaml_file.write("      model: %s\n" % case.MeshName)
    yaml_file.write("      datatype: ExodusII\n")
    yaml_file.write("      var_name: ObjectId\n")
    yaml_file.write("      width: 16cm\n")
    yaml_file.write("      axes: True\n")
    if case.IgnoredBlockKeys:
        yaml_file.write("      ignore_blocks: [%s]\n" % ','.join(case.IgnoredBlockKeys))

    for frag_k, frag_v in fragments.get(sect_index, {}).items():
        insert_fragment(yaml_file, frag_k, frag_v, chap_index, sect_index, backend)

    if verbosity > verbosity_levels.get("terse"):
        # Create mesh blocks section
        sect_index += 1
        yaml_file.write("  - n: section\n")
        yaml_file.write("    title: Mesh Blocks\n")
        yaml_file.write("    string: This section provides a description of all blocks contained in the mesh.\n")
        yaml_file.write("    sections:\n")
        yaml_file.write("    - n: aggregate\n")
        yaml_file.write("      name: show_all_blocks_with_edges\n")
        yaml_file.write("      datatype: ExodusII\n")
        yaml_file.write("      model: %s\n" % case.MeshName)
        if case.DeckType:
            yaml_file.write("      datatype: %s\n" % case.DeckType)
            yaml_file.write("      deckfile: %s\n" % case.DeckRoot)
        if case.LogFile:
            yaml_file.write("      logfile: %s\n" % case.LogFile)
        yaml_file.write("      width: 10cm\n")
        yaml_file.write("      histogram_width: 14cm\n")
        yaml_file.write("      axes: True\n")
        if case.IgnoredBlockKeys:
            yaml_file.write("      ignore_blocks: [%s]\n" % ','.join(case.IgnoredBlockKeys))
        for frag_k, frag_v in fragments.get(sect_index, {}).items():
            insert_fragment(yaml_file, frag_k, frag_v, chap_index, sect_index, backend)


def insert_solution_chapter(yaml_file, case, solution, verbosity_levels, verbosity, chap_index, backend):
    """ Insert solution chapter into YAML structure file
    """

    # Retrieve chapter fragments and initialize section index
    fragments = case.Fragments.get(chap_index, {})
    sect_index = 0

    # Determine name to be reported based on file/partition type
    name = os.path.splitext(
        solution.Name)[0] if "partition" in solution.Type else solution.Name

    # Create YAML entries
    yaml_file.write("# Solution chapter\n")
    yaml_file.write("- n: chapter\n")
    yaml_file.write("  title: %s Solution\n" % solution.Method.title())
    yaml_file.write("  latex: 'This chapter describes the {} solution found in the following directory: {}.'\n".format(
        backend.generate_text(solution.Method),
        backend.generate_text(case.RealDataDir.replace('\\', '/'))))
    yaml_file.write("  sections:\n")
    for frag_k, frag_v in fragments.get(sect_index, {}).items():
        insert_fragment(yaml_file, frag_k, frag_v, chap_index, sect_index, backend)

    # Create new section
    sect_index += 1
    yaml_file.write("  - n: section\n")
    yaml_file.write("    title: Meta-Data\n")
    yaml_file.write("    latex: 'This section lists the meta-information properties found in {}.'\n".format(
        backend.generate_text(name)))
    yaml_file.write("    sections:\n")
    for frag_k, frag_v in fragments.get(sect_index, {}).items():
        insert_fragment(yaml_file, frag_k, frag_v, chap_index, sect_index, backend)
    yaml_file.write("    - n: meta\n")
    yaml_file.write("      datatype: %s\n" % solution.Type)
    yaml_file.write("      dataset: %s\n" % solution.Name)
    yaml_file.write("      verbosity: 0\n")


def insert_stand_alone_chapter(yaml_file, case, chap_index, backend):
    """ Insert stand-alone images and text chapter into YAML structure file
    """

    yaml_file.write("# Third-party artifacts chapter\n")
    yaml_file.write("- n: chapter\n")
    yaml_file.write("  title: Stand-Alone Artifacts\n")
    yaml_file.write(
        "  latex: 'This chapter integrates all standalone images and text fragments found in the following directory: {}.'\n".format(
            backend.generate_text(case.RealDataDir.replace('\\', '/'))))
    yaml_file.write("  sections:\n")
    if case.Images:
        yaml_file.write("  - n: section\n")
        yaml_file.write("    title: PNG Images\n")
        yaml_file.write(
            "    latex: 'This section shows all stand-alone PNG images found in the following directory: {}.'\n".format(
                backend.generate_text(case.RealDataDir.replace('\\', '/'))))
        yaml_file.write("    sections:\n")
        for img in sorted(case.Images):
            print("[Explorator] Including image {}".format(
                img))
            yaml_file.write("    - n: figure\n")
            yaml_file.write("      arguments:\n")
            yaml_file.write("        width: 12cm\n")
            yaml_file.write("        figure_file: %s\n" % img)
            img = img.replace(os.path.sep, '/')
            yaml_file.write("        caption_string: 'File {}'\n".format(
                backend.generate_text(img)))
            yaml_file.write("        label: 'f:%s'\n" % img)
    if case.TextFiles:
        yaml_file.write("  - n: section\n")
        yaml_file.write("    title: Text Fragments\n")
        yaml_file.write(
            "    latex: 'This section shows all stand-alone text fragments found in the following directory: {}.'\n".format(
                backend.generate_text(case.RealDataDir.replace('\\', '/'))))
        yaml_file.write("    sections:\n")
        for txt in sorted(case.TextFiles):
            print("[Explorator] Including text fragment {}".format(
                txt))
            yaml_file.write("    - n: paragraph\n")
            yaml_file.write("      include: %s\n" % os.path.join(case.DataDir, txt))


def generate_structure_file(parameters, case):
    """ Generate YAML structure file from discovered data
    """

    # Open provided structure file
    dst_file = os.path.join(parameters.OutputDir, parameters.StructureFile)
    with open(dst_file, 'w') as yaml_file:
        # Save ARG version on top of structure file
        yaml_file.write("ARG version: {}\n\n".format(parameters.Version))

        # Generate title directive if necessary
        if not parameters.Title and case.DeckType:
            yaml_file.write("# Title directive\n")
            yaml_file.write("title:\n")
            yaml_file.write("  datatype: %s\n" % case.DeckType)
            yaml_file.write("  dataset: %s\n" % case.DeckRoot)
            yaml_file.write("\n")

        # Insert chapters
        yaml_file.write("# Chapters\n")
        yaml_file.write("chapters:\n")
        yaml_file.write("\n")

        # Insert introduction chapter
        chapter_index = 0
        insert_introduction_chapter(yaml_file,
                                    case,
                                    chapter_index,
                                    parameters.Backend)
        chapter_index += 1

        # Insert CAD chapter geometry files are present
        if case.GeometryFiles:
            insert_CAD_chapter(yaml_file,
                               case,
                               parameters.VerbosityLevels,
                               parameters.Verbosity,
                               chapter_index,
                               parameters.Backend)
            chapter_index += 1
        elif not case.GeometryFiles and case.Mappings:
            print("*  ERROR: missing CAD geometry files. Ignoring mappings between CAD and FEM. Exiting. ")
            sys.exit(1)

        # Insert mesh chapter if one was found
        if case.MeshType:
            insert_mesh_chapter(yaml_file,
                                case,
                                parameters.VerbosityLevels,
                                parameters.Verbosity,
                                chapter_index,
                                parameters.Backend)
            chapter_index += 1

        # Insert solution chapter if at least one was found
        for s in case.Solutions:
            insert_solution_chapter(yaml_file,
                                    case,
                                    s,
                                    parameters.VerbosityLevels,
                                    parameters.Verbosity,
                                    chapter_index,
                                    parameters.Backend)
            chapter_index += 1

        # Include third-party artifacts if available
        if parameters.Verbosity > 1 and (
                case.Images or case.TextFiles):
            insert_stand_alone_chapter(yaml_file,
                                       case,
                                       chapter_index,
                                       parameters.Backend)
            chapter_index += 1

        # Append with analyst authored results section structure file is exists
        if parameters.StructureEndFile:
            yaml_file.write(open(parameters.StructureEndFile, 'r').read())

        print("[Explorator] Generated structure file {} including {} chapters".format(
            parameters.StructureFile,
            chapter_index))


if __name__ == '__main__':
    """Main artifact explorator routine
    """

    main(Types, ARG_VERSION)



# HEADER
#                     arg/Aggregation/argAggregatorBase.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
# HEADER

import abc
import os

from arg.DataInterface import argDataInterfaceBase
from arg.Generation import argVTK

class argAggregatorBase:
    """An aggregation abstract base class
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, b, r):

        # A backend is required
        try:
            from arg.Backend import argBackendBase
            assert isinstance(b, argBackendBase.argBackendBase)
            self.Backend = b
        except:
            print("*  WARNING: could not instantiate an aggregator: a backend base is required but a {} was provided".format(
                type(b)))

        # Request parameters are required
        try:
            assert isinstance(r, dict)
            self.RequestParameters = r
        except:
            print("*  WARNING: could not instantiate an aggregator: a request parameters dict is required but a {} was provided".format(
                type(r)))


    def get_backend(self):
        """ Return backend
        """

        return self.Backend


    def get_data_interfaces(self):
        """ Return data interface
        """

        return self.DataInterfaces


    def add_data_interface(self, di):
        """ Set data interface
        """

        if isinstance(di, argDataInterfaceBase.argDataInterfaceBase):
            self.DataInterfaces.append(di)
        else:
            print("*  WARNING: attempted to add incompatible {} to list of data interfacestype. Ignoring it.".format(
                type(di)))


    def show_mesh_surface(self, data, file_name, do_clip=False):
        """A convenience for a request common to several aggregators
        """

        # Create artifact generator variable
        variable = argVTK.argVTKAttribute(data, self.RequestParameters.get(
            "time_step", -1))

        # Execute artifact generator
        print("[argAggregatorBase] Calling argVTK generator for", file_name)
        output_base_name, caption = argVTK.four_surfaces(
            self.Backend.Parameters,
            self.RequestParameters,
            data,
            variable,
            file_name,
            do_clip)

        # Include figure in report
        self.Backend.add_figure({
            "figure_file": output_base_name + ".png",
            "caption_string": caption,
            "width": self.RequestParameters.get("width", "12cm")})


    @abc.abstractmethod
    def aggregate(self):
        """Decide which aggregation operation is to be performed
        """

        pass




#HEADER
#                         arg/Applications/ARG.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import distutils.spawn
import getopt
import os
import shutil
import sys
import time
import yaml

from arg import __version__
from arg.Common.argReportParameters import argReportParameters
from arg.Applications import Explorator
from arg.Applications.argGenerator import argGenerator

ARG_VERSION = __version__

# Import ARG modules
if not __package__:
    sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
else:
    sys.path.append("..")

# Load supported types
common_dir = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(common_dir, "../Common/argTypes.yml"),
          'r',
          encoding="utf-8") as t_file:
    Types = yaml.safe_load(t_file)


class Runner:
    """A class to describe ARG runner parameters
    """

    def __init__(self, version=None):
        """Class constructor
        """

        # Default values of variables that must exist
        self.Explore = False
        self.Generate = False
        self.Assemble = False
        self.ParametersFile = None
        self.Parameters = None
        self.Version = version
        self.LatexProcessor = None
        self.TexFile = None

    @staticmethod
    def usage():
        """Provide online help
        """

        print("Usage:")
        print("\t [-h]                      Help: print this message and exit")
        print("\t [-e]                      run Explorator then Assembler")
        print("\t [-g]                      run Generator then Assembler")
        print("\t [-E]                      run Explorator")
        print("\t [-G]                      run Generator")
        print("\t [-A]                      run Assembler")
        print("\t [-p <parameters file>]    name of parameters file")
        print("\t [-l <LaTeX processor>]    name of LaTeX processor")
        print("\t [-t]                      generate just .tex file")
        sys.exit(0)

    def parse_line(self, default_parameters_filename, types=None):
        """Parse command line and fill artifact parameters
        """

        # Initiate caller as empty string by default
        caller = ""

        # Try to hash command line with respect to allowable flags
        try:
            opts, args = getopt.getopt(sys.argv[1:], "egEGAp:l:t")
        except getopt.GetoptError:
            self.usage()
            sys.exit(1)

        # Parse arguments and assign corresponding variables
        for o, a in opts:
            if o == "-h":
                self.usage()
                sys.exit(0)
            elif o == "-e":
                self.Assemble = True
                self.Explore = True
                caller = "Explorator"
            elif o == "-g":
                self.Assemble = True
                self.Generate = True
                caller = "Generator"
            elif o == "-E":
                self.Explore = True
                caller = "Explorator"
            elif o == "-G":
                self.Generate = True
                caller = "Generator"
            elif o == "-A":
                self.Assemble = True
                caller = "Assembler"
            elif o == "-p":
                self.ParametersFile = a
            elif o == "-l":
                self.LatexProcessor = a
            elif o == "-t":
                self.TexFile = True

        # Inform user if missing argument
        if not self.ParametersFile:
            print("*  WARNING: no parameters file name provided; "
                  "using '{}' by default.".format(default_parameters_filename))
            self.ParametersFile = default_parameters_filename

        # Create parameters
        self.Parameters = argReportParameters("ARG",
                                              self.ParametersFile,
                                              self.Version,
                                              types,
                                              self.LatexProcessor)
        self.Parameters.TexFile = self.TexFile

        # Populate attributes based on parameters file content
        return (self.Parameters.check_parameters_file()
                and self.Parameters.parse_parameters_file()
                and self.Parameters.check_parameters(caller))


    def run(self):
        """Run ARG applications based on parsed values
        """

        # Copy Explorator output file if exists
        concatenateStructureFile = False
        if self.Explore and os.path.isfile("{}.yml".format(self.Parameters.StructureFile)):
            concatenateStructureFile = True
            shutil.copyfile(os.path.realpath("{}.yml".format(self.Parameters.StructureFile)),
                            os.path.realpath("{}_tmp.yml".format(self.Parameters.StructureFile)))

        # Instantiate generator
        generator = argGenerator(self.Parameters)
            
        # Explore when required
        if self.Explore:
            Explorator.execute(self.Parameters)

        # Generate when required
        if self.Generate:
            generator.generate_artefacts()

        # Concatenate existing structure file content with generated content
        if concatenateStructureFile:
            with open(self.Parameters.StructureFile, 'w') as newFile:
                with open(self.Parameters.StructureFile, 'r') as existingFile:
                    newFile.write(existingFile.read())

        # Assemble when required
        if self.Assemble:
            generator.assemble_report()


def main(types, version=None):
    """ ARG main method
    """

    # Start stopwatch
    t_start = time.time()

    # Print startup information
    sys_version = sys.version_info
    print("[ARG] ### Started with Python {}.{}.{}".format(
        sys_version.major,
        sys_version.minor,
        sys_version.micro))

    # Retrieve default parameters filename from provided types dict
    default_parameters_filename = types.get("DefaultParametersFile")

    # Instantiate runner and execute it
    runner = Runner(version)
    runner.parse_line(default_parameters_filename, types)
    runner.run()

    # End stopwatch
    dt = time.time() - t_start

    # If this point is reached everything went fine
    print("[ARG] Process completed in {} seconds. ###".format(dt))


if __name__ == '__main__':
    """Main ARG routine
    """

    main(Types, ARG_VERSION)



#HEADER
#                  arg/DataInterface/argDataInterface.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import importlib


class argDataInterface:

    @staticmethod
    def factory(data_type, database_name, *parameters):
        """Produce the necessary concrete data interface instance
        """

        # Initialize return object
        ret_object = None

        # Collection of co-located key=value text files
        if data_type == "key-value":
            try:
                argKeyValueFilesReader = getattr(
                    importlib.import_module("arg.DataInterface.argKeyValueFilesReader"),
                    "argKeyValueFilesReader")
                ret_object = argKeyValueFilesReader(database_name, *parameters)
            except:
                print("*  WARNING: could not import module argKeyValueFilesReader. Ignoring it.")

        # HDF5 file
        elif data_type == "HDF5":
            try:
                argHDF5Reader = getattr(
                    importlib.import_module("arg.DataInterface.argHDF5Reader"),
                    "argHDF5Reader")
                ret_object = argHDF5Reader(database_name, *parameters)
            except:
                print("*  WARNING: could not import module argHDF5Reader. Ignoring it.")

        # ExodusII file or partition
        elif data_type == "ExodusII":
            try:
                argVTKExodusReader = getattr(
                    importlib.import_module("arg.DataInterface.argVTKExodusReader"),
                    "argVTKExodusReader")
                ret_object = argVTKExodusReader(database_name, *parameters)
            except:
                print("*  WARNING: could not import module argVTKExodusReader. Ignoring it.")

        # STL file
        elif data_type == "vtkSTL":
            try:
                argVTKSTLReader = getattr(
                    importlib.import_module("arg.DataInterface.argVTKSTLReader"),
                    "argVTKSTLReader")
                ret_object = argVTKSTLReader(database_name, *parameters)
            except:
                print("*  WARNING: could not import module argVTKSTLReader. Ignoring it.")

        # Return instantiated object
        if ret_object:
            print("[argDataInterface] Instantiated {} reader for {}".format(
                data_type,
                database_name))
        else:
            print("[argDataInterface] Could not instantiate {} reader for {}".format(
                data_type,
                database_name))
        return ret_object



#HEADER
#                        arg/Generation/argPlot.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
#HEADER

import math
import os
import string
import sys

import matplotlib
import matplotlib.pyplot
import matplotlib.pylab as mpl
import numpy as np
import yaml

from arg.Common.argMultiFontStringHelper import argMultiFontStringHelper
from arg.DataInterface.argDataInterface import argDataInterface

matplotlib.use("Agg")

# Available colors
argPlot_colors = [
    "blue", "magenta", "black", "green", "cyan ",
    "pink", "yellow", "red", "gray", "darkgray",
    "lightgray", "brown", "lime", "olive", "orange",
    "purple", "teal", "violet"]

# Available styles
argPlot_styles = ["solid", "dashed", "dashdot"]

# Available mathematical expressions
argPlot_expressions = dict([(f, getattr(math, f)) for f in [
    "sqrt", "pow",
    "ceil", "floor", "fabs",
    "pi", "cos", "sin", "tan", "acos",
    "asin", "atan", "atan2",
    "e", "cosh", "sinh", "tanh",
    "exp", "log", "log10"]])

# Load supported types
common_dir = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(common_dir, "../Common/argTypes.yml"),
          'r',
          encoding="utf-8") as t_file:
    supported_types = yaml.safe_load(t_file)

# Retrieve supported MatPlotLib-based plottings
plotting_functions = supported_types.get(
    "PlottingFunctions")

# Figure extension constants
constant_shift = .05
decrease_factor = .95
increase_factor = 1.05


def safely_evaluate_expression(expr, x):
    """Evaluate expression at x only allowing for explicitly supported expressions
    """

    # Add variable x into dictionary of safe expressions
    argPlot_expressions['x'] = x

    # Evaluate expression at x and return if no exception
    try:
        y = eval(expr, {"__builtins__": None}, argPlot_expressions)
    except:
        y = None

    # Return value
    return y


def make_base_name(title, fct_label, var_x_name, var_y_name, suffix):
    """ Assemble image name from title, function label, variable names, and suffix
    """

    # Pack title and variable names
    t_packed = '-'.join(title.split())
    x_packed = '-'.join(var_x_name.split())
    y_packed = '-'.join(var_y_name.split())

    # Create valid image file name parts
    s = '_'.join([t_packed, fct_label, x_packed, y_packed, suffix])
    valid_chars = "-_.() {}{}".format(
        string.ascii_letters,
        string.digits)
    base_name = ''.join(c for c in s if c in valid_chars)

    # Replace spaces and parentheses with underscores
    base_name = base_name.replace(' ', '_').replace('(', '_').replace(')', '_')

    # Prevent hidden or directory-like base names
    if base_name.startswith('.'):
        base_name = "_{}".format(base_name)

    # Return non-empty base name
    return base_name if base_name else '_'


def create_caption(backend,
                   prefix,
                   provenance=None,
                   var_name=None,
                   var_x_name=None,
                   suffix=None,
                   function=None,
                   model=None,
                   material=None):
    """Create caption depending on concrete backend
    """

    # Create a multi-font caption when a backend is provided
    if backend:
        # Instantiante multi-font string for given backend
        caption_string = argMultiFontStringHelper(backend)

        if prefix:
            caption_string.append(
                prefix[0].upper() + prefix[1:] + " plot", "default")
            if var_name or provenance:
                caption_string.append(" of ", "default")

            # Append variable name when available
            if var_name:
                caption_string.append(
                    "{}".format(var_name), "typewriter")

            # Connect variable name to provenance name when both are provided
            if var_name and provenance:
                caption_string.append(" in ", "default")

            # Append provenance name when provided
            if provenance:
                caption_string.append(provenance, "typewriter")

        else:
            if not var_x_name and not var_name:
                caption_string.append("Curve", "default")
            elif not var_name:
                caption_string.append("{}".format(var_x_name), "typewriter")
                caption_string.append(" curve", "default")
            elif not var_x_name:
                caption_string.append("{}".format(var_name), "typewriter")
                caption_string.append(" curve", "default")
            else:
                caption_string.append("{}".format(var_name), "typewriter")
                caption_string.append(" vs. ", "italic")
                caption_string.append("{}".format(var_x_name), "typewriter")
                caption_string.append(" curve", "default")

            # Append function when available
            if function:
                caption_string.append(" of", "default")
                caption_string.append(" {}".format(function), "typewriter")
                caption_string.append(" function", "default")

            # Append model when available
            if model:
                caption_string.append(" of", "default")
                caption_string.append(" {}".format(model), "typewriter")
                caption_string.append(" model", "default")

            # Append material when available
            if material:
                caption_string.append(" for", "default")
                caption_string.append(" {}".format(material), "typewriter")
                caption_string.append(" material", "default")

        # Close caption string
        if suffix:
            caption_string.append(suffix, "default")
        caption_string.append('.', "default")

    # Otherwise create a plain string
    else:
        caption_string = prefix[0].upper() + prefix[1:] + " plot"
        if var_name or provenance:
            caption_string += " of "

        # Append variable name when available
        if var_name:
            caption_string += "{}".format(var_name)

        # Connect variable name to provenance name when both are provided
        if var_name and provenance:
            caption_string += " in "

        # Append provenance name when provided
        if provenance:
            caption_string += provenance

        # Close caption string
        if suffix:
            caption_string += suffix
        caption_string += '.'

    # Return caption string
    return caption_string


def compute_aspect_ratio(x, y, stretch_factor):
    """Compute aspect ratio of chart from points coordinates and stretch factor
    """

    # Storage for [range_x, range_y]
    ranges = []

    # Ensure that each axis is not degenerate
    for axis in (x, y):
        # Initialize with axis extrema
        axis_min, axis_max = min(axis), max(axis)

        # Check whether axis needs adjustment
        if axis_min == axis_max:
            if not axis_min:
                # Extend axis by a constant value
                axis_min, axis_max = -constant_shift, constant_shift
            else:
                # Extend axis by scalar factors
                axis_min *= increase_factor if np.sign(axis_min) < 0 else decrease_factor
                axis_max *= decrease_factor if np.sign(axis_max) < 0 else increase_factor

        # Store possibly adjusted range
        ranges.append(axis_max - axis_min)

    # Compute and return aspect ratio
    return ranges[0] / (stretch_factor * ranges[1])


def time(parameters, plot_params):
    """Create plot of specified variable versus time
    """

    # Retrieve plot title
    try:
        title = plot_params["title"]
    except:
        return None, None, "** ERROR: no title provided for time plot request. Ignoring it."

    # Retrieve list of files
    try:
        files = plot_params["files"]
    except:
        return None, None, "** ERROR: no labels specified for time plot request. Ignoring it."

    # Retrieve variable name
    try:
        var_name = plot_params["var_name"]
    except:
        return None, None, "** ERROR: no variable name specified for time plot request. Ignoring it."

    # Try to retrieve parameters
    function = plot_params.get("function")
    model = plot_params.get("model")
    material = plot_params.get("material")
    fct_temp = plot_params.get("temperature")

    # Determine caption string depending on existence of temperature
    str_caption = " at temperature = {}".format(fct_temp) if fct_temp else None

    # Assemble image name
    output_base_name = make_base_name(
        title,
        '-'.join([f for f in files]),
        '',
        var_name,
        "time_plot")
    image_full_name = os.path.join(parameters.OutputDir, output_base_name) + ".png"

    # Create caption
    if function or model or material:
        prefix = None
    else:
        prefix = "time"
    caption = create_caption(
        parameters.Backend,
        prefix,
        title,
        var_name,
        None,
        str_caption,
        function,
        model,
        material)

    # Generate image only if not already present
    if os.path.isfile(image_full_name):
        return output_base_name, caption, None

    # Retrieve list of labels
    try:
        labels = plot_params["labels"]
    except:
        return None, None, "** ERROR: no labels specified for time plot request. Ignoring it."

    # Sanity check: labels must correspond to files
    if len(files) != len(labels):
        return None, None, "** ERROR: files and labels mismatch in time plot request. Ignoring it."

    # Retrieve value index
    val_index = int(plot_params.get("val_index", 0))

    # Initialize min/max containers
    t_min, t_max, v_min, v_max = [], [], [], []

    # Create one plot per file
    plots = []
    for i, f in enumerate(files):
        # Get data through ExodusII interface
        data = argDataInterface.factory(
            "ExodusII",
            os.path.join(parameters.DataDir, f),
            var_name)

        # Retrieve available times and go to next plot if empty
        avail_times = data.get_available_times()
        if not avail_times:
            continue

        # Update time ranges
        t_min.append(min(avail_times))
        t_max.append(max(avail_times))

        # Iterate over time steps
        values = []
        for t in range(0, len(avail_times)):
            # Retrieve list of variable values at given time
            time_slice = data.get_variable_time_slice(t, var_name, True)
            values.append(time_slice[val_index])

        # Iterate over time steps
        v_min.append(min(values))
        v_max.append(max(values))
        plots.append([
            avail_times,
            values,
            labels[i],
            argPlot_colors[i % len(argPlot_colors)],
            argPlot_styles[i % len(argPlot_styles)]])

    # If no available times were found, do not do anything
    if not t_min:
        return None, None, "** ERROR: no available times found for time plot request. Ignoring it."

    # Create chart
    mpl.rc("text", usetex=True)
    fig, ax = matplotlib.pyplot.subplots()

    # Add title
    if parameters.BackendType == "LaTeX":
        ax.set_title(
            r"\large\texttt{{{}}}".format(title.replace('_', r"\_")),
            fontweight="bold")
    else:
        ax.set_title(
            r"{}".format(title.replace('_', r"\_")),
            fontweight="bold")

    # Retrieve aspect ratio of figure
    try:
        ax.set_aspect(float(plot_params["fig_ratio"]))
    except:
        ax.set_aspect("auto")

    # Set extrema
    ax.set_xlim([min(t_min), max(t_max)])
    ax.set_ylim([min(v_min), max(v_max)])

    # Set labels
    ax.set_xlabel("Time (sec)")
    try:
        ax.set_ylabel(r"large %s" % plot_params["ylabel"].replace('_', r"\_"))
    except:
        ax.set_ylabel(r"large %s" % var_name)

    # Add all plots
    for [t, v, l, c, s] in plots:
        ax.plot(t, v, label=l, color=c, ls=s, lw=.5)
    ax.legend(loc="lower right", fontsize=8)

    # Export chart to PNG file
    try:
        fig.savefig(image_full_name, bbox_inches="tight", pad_inches=0, transparent=True)
        fig.clf()
        matplotlib.pyplot.close(fig)
    except:
        return None, None, "** ERROR: plot request could not save created file {} with exception `{}`. Ignoring it.".format(
            output_base_name,
            sys.exc_info())

    # If this point was reached everything went well
    return output_base_name, caption, None


def xy(parameters, plot_params):
    """Create continuous piecewise linear plot joining provided x,y data
    """

    # Retrieve plot title
    try:
        title = plot_params["title"]
    except:
        return None, None, "** ERROR: no title provided for xy plot request. Ignoring it."

    # Retrieve plot label
    try:
        fct_label = plot_params["label"]
    except:
        return None, None, "** ERROR: no label specified for xy plot request. Ignoring it."

    # Set variable column indices
    var_x_col = plot_params.get("var_x_column", 0)
    var_y_col = plot_params.get("var_y_column", 1)

    # Retrieve variable names
    try:
        var_x_name = plot_params["var_x_name"].replace('_', r"\_")
    except:
        var_x_name = "column " + "%s" % var_x_col
    try:
        var_y_name = plot_params["var_y_name"].replace('_', r"\_")
    except:
        var_y_name = "column " + "%s" % var_y_col

    # Try to retrieve parameters
    function = plot_params.get("function")
    model = plot_params.get("model")
    material = plot_params.get("material")
    fct_temp = plot_params.get("temperature")

    # Determine caption string depending on existence of temperature
    str_caption = " at temperature = {}".format(fct_temp) if fct_temp else None

    # Assemble image name
    output_base_name = make_base_name(
        title,
        fct_label,
        var_x_name,
        var_y_name,
        "xy_plot")
    image_full_name = os.path.join(parameters.OutputDir, output_base_name) + ".png"

    # Create caption
    if function or model or material:
        prefix = None
    else:
        prefix = "piecewise linear"
    caption = create_caption(
        parameters.Backend,
        prefix,
        title,
        var_y_name,
        var_x_name,
        str_caption,
        function,
        model,
        material)

    # Generate image only if not already present
    if os.path.isfile(image_full_name):
        return output_base_name, caption, None

    # Retrieve data for x and y variables
    try:
        data = plot_params["data"]
    except:
        return None, None, "** ERROR: no data provided for xy plot request. Ignoring it."
    x = []
    y = []

    # Iterate over data entries
    for row in data:
        # Distinguish between text and numeric data
        if isinstance(row, str):
            # Eliminate commas and split row to extract data
            row = row.strip().replace(',', ' ').split()
        elif not isinstance(row, list):
            return None, None, "** ERROR: unsuited data format for xy plot request. Ignoring it."

        # Fetch first two dimensions
        try:
            x.append(float(row[var_x_col]))
            y.append(float(row[var_y_col]))
        except:
            return None, None, "** ERROR: unsuited data format for xy plot request. Ignoring it."

    # If no available or incorrect data values were found, do not do anything
    if not len(x):
        return None, None, "** ERROR: no available data found for xy plot request. Ignoring it."
    if len(x) != len(y):
        return None, None, "** ERROR: x/y data mismatch for xy plot request. Ignoring it."

    # Create chart
    mpl.rc("text", usetex=True)
    fig, ax = matplotlib.pyplot.subplots()

    # Add title
    if parameters.BackendType == "LaTeX":
        ax.set_title(
            r"\large\texttt{{{}}}".format(fct_label.replace('_', r"\_")),
            fontweight="bold")
    else:
        ax.set_title(
            r"{}".format(fct_label.replace('_', r"\_")),
            fontweight="bold")

    # Set labels
    if var_x_name:
        if parameters.BackendType == "LaTeX":
            ax.set_xlabel(r"\texttt{{{}}}".format(var_x_name))
        else:
            ax.set_xlabel(var_x_name)
    if var_y_name:
        if parameters.BackendType == "LaTeX":
            ax.set_ylabel(r"\texttt{{{}}}".format(var_y_name))
        else:
            ax.set_xlabel(var_y_name)
    matplotlib.pyplot.grid(True)

    # Determine type of marker if any
    marker_type = plot_params.get("marker")

    # Create plot
    ax.set_aspect(compute_aspect_ratio(x, y, plot_params["xyratio"]))
    if isinstance(marker_type, list):
        ax.plot(
            x[:2], y[:2],
            color="blue", ls="solid", lw=1)
        ax.plot(
            x[1], y[1],
            marker=marker_type[0], color="blue", ls="solid", lw=1, ms=4)
        ax.plot(
            x[1:], y[1:],
            color="blue", ls="solid", lw=1)
    else:
        ax.plot(
            x, y,
            marker=marker_type, color="blue", ls="solid", lw=1, ms=2)

    # Export chart to PNG file
    try:
        fig.savefig(image_full_name, bbox_inches="tight", pad_inches=0, transparent=True)
        fig.clf()
        matplotlib.pyplot.close(fig)
    except:
        return None, None, "** ERROR: plot request could not save created file {} with exception `{}`. Ignoring it.".format(
            output_base_name,
            sys.exc_info())

    # If this point was reached everything went well
    return output_base_name, caption, None


def lin_exp(parameters, plot_params):
    """Create continuous linear then exponential plot with given parameters
    """

    # Retrieve plot title
    try:
        title = plot_params["title"]
    except:
        return None, None, "** ERROR: no title provided for lin_exp plot request. Ignoring it."

    # Retrieve plot label
    try:
        fct_label = plot_params["label"]
    except:
        return None, None, "** ERROR: no label specified for lin_exp plot request. Ignoring it."

    # Set variable names
    var_x_name = plot_params.get("var_x_name", '')
    var_y_name = plot_params.get("var_y_name", '')

    # Try to retrieve parameters
    function = plot_params.get("function")
    model = plot_params.get("model")
    material = plot_params.get("material")
    fct_temp = plot_params.get("temperature")

    # Determine caption string depending on existence of temperature
    str_caption = " at temperature = {}".format(fct_temp) if fct_temp else None

    # Assemble image name
    output_base_name = make_base_name(
        title,
        fct_label,
        var_x_name,
        var_y_name,
        "lin_exp_plot")
    image_full_name = os.path.join(parameters.OutputDir, output_base_name) + ".png"

    # Create caption
    if function or model or material:
        prefix = None
    else:
        prefix = "linear/exponential"
    caption = create_caption(
        parameters.Backend,
        prefix,
        title,
        var_y_name,
        var_x_name,
        str_caption,
        function,
        model,
        material)

    # Generate image only if not already present
    if os.path.isfile(image_full_name):
        return output_base_name, caption, None

    # If no available or incorrect data values were found, do not do anything
    try:
        data = plot_params["data"]
    except:
        return None, None, "** ERROR: no data provided for lin_exp plot request. Ignoring it."
    if len(data) < 3:
        return None, None, "** ERROR: not enough data found for lin_exp plot request. Ignoring it."

    # Retrieve data
    x = [data[0][0], data[1][0], data[2]]
    y = [data[0][1], data[1][1]]
    lin_slope = x[1] / y[1]
    inv_const = 1. / data[3]
    inv_exp = 1. / data[4]

    # Look for y intercept at given maximum x
    dy = (y[1] - y[0]) / 100.
    x_max, y_max = x[1], y[1]
    while x_max < x[2]:
        y_max += dy
        x_max = ((y_max - y[1]) * inv_const) ** inv_exp + y_max * lin_slope
    y.append(y_max - dy)

    # Compute exponential data
    exp_y_range = np.linspace(y[1], y[2], 100)
    exp_x_vals = ((exp_y_range - y[1]) * inv_const) ** inv_exp + exp_y_range * lin_slope

    # Create chart
    mpl.rc("text", usetex=True)
    fig, ax = matplotlib.pyplot.subplots()

    # Add title
    if parameters.BackendType == "LaTeX":
        ax.set_title(
            r"\large\texttt{{{}}}".format(fct_label.replace('_', r"\_")),
            fontweight="bold")
    else:
        ax.set_title(
            r"{}".format(fct_label.replace('_', r"\_")),
            fontweight="bold")

    # Set labels
    if var_x_name:
        if parameters.BackendType == "LaTeX":
            ax.set_xlabel(r"\texttt{{{}}}".format(var_x_name))
        else:
            ax.set_xlabel(var_x_name)
    if var_y_name:
        if parameters.BackendType == "LaTeX":
            ax.set_ylabel(r"\texttt{{{}}}".format(var_y_name))
        else:
            ax.set_xlabel(var_x_name)
    matplotlib.pyplot.grid(True)

    # Create plot
    ax.set_aspect(compute_aspect_ratio(x, y, plot_params["xyratio"]))
    ax.plot(
        x[:-1], y[:-1],
        color="blue", ls="solid", lw=1)
    ax.plot(
        x[1], y[1], marker='D',
        color="blue", ls="solid", lw=1, ms=4)
    ax.plot(
        exp_x_vals, exp_y_range,
        color="blue", ls="solid", lw=1)

    # Export chart to PNG file
    try:
        fig.savefig(image_full_name, bbox_inches="tight", pad_inches=0, transparent=True)
        fig.clf()
        matplotlib.pyplot.close(fig)
    except:
        return None, None, "** ERROR: plot request could not save created file {} with exception `{}`. Ignoring it.".format(
            output_base_name,
            sys.exc_info())

    # If this point was reached everything went well
    return output_base_name, caption, None


def analytic(parameters, plot_params):
    """Create plot from analytic definition with given parameters
    """

    # Retrieve plot title
    try:
        title = plot_params["title"]
    except:
        return None, None, "** ERROR: no title provided for analytic plot request. Ignoring it."

    # Retrieve plot label
    try:
        fct_label = plot_params["label"]
    except:
        return None, None, "** ERROR: no label specified for analytic plot request. Ignoring it."

    # Set variable names
    var_x_name = plot_params.get("var_x_name", '')
    var_y_name = plot_params.get("var_y_name", '')

    # Try to retrieve parameters
    function = plot_params.get("function")
    model = plot_params.get("model")
    material = plot_params.get("material")
    fct_temp = plot_params.get("temperature")

    # Determine caption string depending on existence of temperature
    str_caption = " at temperature = {}".format(fct_temp) if fct_temp else None

    # Assemble image name
    output_base_name = make_base_name(
        title,
        fct_label,
        var_x_name,
        var_y_name,
        "analytic_plot")
    image_full_name = os.path.join(parameters.OutputDir, output_base_name) + ".png"

    # Create caption
    if function or model or material:
        prefix = None
    else:
        prefix = "piecewise linear"
    caption = create_caption(
        parameters.Backend,
        prefix,
        title,
        var_y_name,
        var_x_name,
        str_caption,
        function,
        model,
        material)

    # Generate image only if not already present
    if os.path.isfile(image_full_name):
        return output_base_name, caption, None

    # If no available or incorrect data values were found, do not do anything
    try:
        data = plot_params["data"]
    except:
        return None, None, "** ERROR: no data provided for analytic plot request. Ignoring it."
    if len(data) < 2:
        return None, None, "** ERROR: not enough data found for analytic plot request. Ignoring it."

    # Retrieve data
    f = data[0]
    [x1, x2] = [float(x) for x in data[1]]
    if x1 > x2:
        x1, x2 = x2, x1

    # Initialize arrays of values
    y1 = safely_evaluate_expression(f, x1)
    x, y = [x1], [y1]

    # Build array of values
    dx = (x2 - x1) / 100.
    y_min, y_max = y1, y1
    while x1 < x2:
        x1 += dx
        x.append(x1)
        y1 = safely_evaluate_expression(f, x1)
        y.append(y1)
        if y1 < y_min:
            y_min = y1
        elif y1 > y_max:
            y_max = y1

    # Create chart
    mpl.rc("text", usetex=True)
    fig, ax = matplotlib.pyplot.subplots()

    # Add title
    if parameters.BackendType == "LaTeX":
        ax.set_title(
            r"\large\texttt{{{}}}".format(fct_label.replace('_', r"\_")),
            fontweight="bold")
    else:
        ax.set_title(
            r"{}".format(fct_label.replace('_', r"\_")),
            fontweight="bold")

    # Set labels
    if var_x_name:
        if parameters.BackendType == "LaTeX":
            ax.set_xlabel(r"\texttt{{{}}}".format(var_x_name))
        else:
            ax.set_xlabel(var_x_name)
    if var_y_name:
        if parameters.BackendType == "LaTeX":
            ax.set_ylabel(r"\texttt{{{}}}".format(var_y_name))
        else:
            ax.set_xlabel(var_y_name)
    matplotlib.pyplot.grid(True)

    # Create plot
    ax.set_aspect(compute_aspect_ratio(x, y, plot_params["xyratio"]))
    ax.plot(x, y, color="blue", ls="solid", lw=1)

    # Export chart to PNG file
    try:
        fig.savefig(image_full_name, bbox_inches="tight", pad_inches=0, transparent=True)
        fig.clf()
        matplotlib.pyplot.close(fig)
    except:
        return None, None, "** ERROR: plot request could not save created file {} with exception `{}`. Ignoring it.".format(
            output_base_name,
            sys.exc_info())

    # If this point was reached everything went well
    return output_base_name, caption, None


def histogram(parameters, plot_params):
    """Create histogram plot of provided map data
    """

    # Retrieve plot title
    try:
        title = plot_params["title"]
    except:
        return None, None, "** ERROR: no title provided for histogram plot request. Ignoring it."

    # Retrieve variable names
    try:
        var_x_name = plot_params["var_x_name"].replace('_', r"\_")
    except:
        return None, None, "** ERROR: no variable name provided for histogram plot request. Ignoring it."
    try:
        var_y_name = plot_params["var_y_name"].replace('_', r"\_")
    except:
        var_y_name = "count"

    # Assemble image name
    output_base_name = make_base_name(
        title,
        '',
        var_x_name,
        '',
        "histogram_plot")
    image_full_name = os.path.join(parameters.OutputDir, output_base_name) + ".png"

    # Create caption
    caption = create_caption(
        parameters.Backend,
        "histogram",
        title,
        var_y_name)

    # Generate image only if not already present
    if os.path.isfile(image_full_name):
        return output_base_name, caption, None

    # Retrieve data for x and y variables
    try:
        data = plot_params["data"]
    except:
        return None, None, "** ERROR: no data provided for histogram plot request. Ignoring it."

    # Iterate over data map
    x, w = [], []
    for k, v in sorted(data.items()):
        x.append(float(k))
        w.append(v)

    # If no available or incorrect data values were found, do not do anything
    if not len(x):
        return None, None, "** ERROR: no available data found for histogram plot request. Ignoring it."

    # Create chart
    mpl.rc("text", usetex=True)
    fig, ax = matplotlib.pyplot.subplots()

    # Add title
    if parameters.BackendType == "LaTeX":
        ax.set_title(
            r"\large\texttt{{{}}}".format(title.replace('_', r"\_")),
            fontweight="bold")
    else:
        ax.set_title(
            r"{}".format(title.replace('_', r"\_")),
            fontweight="bold")

    # Set labels
    ax.set_xlabel(var_x_name)
    ax.set_ylabel(var_y_name)
    matplotlib.pyplot.grid(True)

    # Determine type of marker if any
    marker_type = plot_params.get("marker")

    # Create histogam if it makes sense
    n, bins, patches = ax.hist(x, 15, weights=w, facecolor="blue", alpha=.5)
    ax.set_aspect(compute_aspect_ratio(x, [0., max(n)], plot_params["xyratio"]))

    # Export chart to PNG file
    try:
        fig.savefig(image_full_name, bbox_inches="tight", pad_inches=0, transparent=True)
        fig.clf()
        matplotlib.pyplot.close(fig)
    except:
        return None, None, "** ERROR: plot request could not save created file {} with exception `{}`. Ignoring it.".format(
            output_base_name,
            sys.exc_info())

    # If this point was reached everything went well
    return output_base_name, caption, None


def constant(parameters, plot_params):
    """Create constant plot joining provided data on [0;1]
    """

    var_x_col = plot_params.get("var_x_column", 0)
    var_y_col = plot_params.get("var_y_column", 1)

    # Retrieve plot title
    try:
        title = plot_params["title"]
    except:
        return None, None, "** ERROR: no title provided for constant plot request. Ignoring it."

    # Retrieve plot label
    try:
        fct_label = plot_params["label"]
    except:
        return None, None, "** ERROR: no label specified for constant plot request. Ignoring it."

    # Set variable column indices
    cst = float(plot_params.get("data", 1)[0])
    x = [0., 1.]
    y = [cst, cst]

    # Retrieve variable names
    try:
        var_x_name = plot_params["var_x_name"].replace('_', r"\_")
    except:
        var_x_name = "column " + "%s" % var_x_col
    try:
        var_y_name = plot_params["var_y_name"].replace('_', r"\_")
    except:
        var_y_name = "column " + "%s" % var_y_col

    # Retrieve temperature
    try:
        fct_temp = plot_params["temperature"]
    except:
        fct_temp = None
    if fct_temp:
        str_caption = " at temperature = {}".format(fct_temp)
    else:
        str_caption = None

    # Assemble image name
    output_base_name = make_base_name(
        title,
        fct_label,
        var_x_name,
        var_y_name,
        "constant_plot")
    image_full_name = os.path.join(parameters.OutputDir, output_base_name) + ".png"

    # Create caption
    caption = create_caption(
        parameters.Backend,
        "constant",
        title,
        var_y_name,
        str_caption)

    # Generate image only if not already present
    if os.path.isfile(image_full_name):
        return output_base_name, caption, None

    # If no available or incorrect data values were found, do not do anything
    if not len(x):
        return None, None, "** ERROR: no available data found for constant plot request. Ignoring it."

    # Create chart
    mpl.rc("text", usetex=True)
    fig, ax = matplotlib.pyplot.subplots()

    # Hide x-axis tick labels
    ax.set_xticklabels([])

    # Add title
    if parameters.BackendType == "LaTeX":
        ax.set_title(
            r"\large\texttt{{{}}}".format(fct_label.replace('_', r"\_")),
            fontweight="bold")
    else:
        ax.set_title(
            r"{}".format(fct_label.replace('_', r"\_")),
            fontweight="bold")

    # Set labels
    if var_x_name:
        if parameters.BackendType == "LaTeX":
            ax.set_xlabel(r"\texttt{{{}}}".format(var_x_name))
        else:
            ax.set_xlabel(var_x_name)
    if var_y_name:
        if parameters.BackendType == "LaTeX":
            ax.set_xlabel(r"\texttt{{{}}}".format(var_y_name))
        else:
            ax.set_xlabel(var_y_name)
    matplotlib.pyplot.grid(True)

    # Determine type of marker if any
    marker_type = plot_params.get("marker")

    # Create plot
    ax.set_aspect(compute_aspect_ratio(x, y, plot_params["xyratio"]))
    if isinstance(marker_type, list):
        ax.plot(
            x[:2], y[:2],
            color="blue", ls="solid", lw=1)
        ax.plot(
            x[1], y[1],
            marker=marker_type[0], color="blue", ls="solid", lw=1, ms=4)
        ax.plot(
            x[1:], y[1:],
            color="blue", ls="solid", lw=1)
    else:
        ax.plot(
            x, y,
            marker=marker_type, color="blue", ls="solid", lw=1, ms=2)
    # Export chart to PNG file
    try:
        fig.savefig(image_full_name, bbox_inches="tight", pad_inches=0, transparent=True)
        fig.clf()
        matplotlib.pyplot.close(fig)
    except:
        return None, None, "** ERROR: plot request could not save created file {} with exception `{}`. Ignoring it.".format(
            output_base_name,
            sys.exc_info())

    # If this point was reached everything went well
    return output_base_name, caption, None


def execute_request(parameters, plot_params):
    """Execute MatPlotLib plotting request to find or create some artifact(s)
    """

    # Bail out if no plot type was specified
    req_name = plot_params.get("type")
    if not req_name:
        print("** ERROR: no plot type was specified in request. Ignoring it.")
        return None, None

    # Proceed with request
    print("[argPlot] Processing {} plot request".format(
        req_name))

    # Try to execute requested visualization function
    for f in plotting_functions:
        if req_name.startswith(f):
            # Execute known plotting function and break out from loop
            output_base_name, caption, warn = globals()[f](
                parameters, plot_params)
            break
    else:
        # Unsupported function was requested
        print("** ERROR: `{}` plotting is not supported. Ignoring it.")
        return None, None

    # Print warning message if any and return accordingly
    if warn:
        print(warn)
        return None, None
    else:
        # Return plot base name and caption when everything went well
        return output_base_name, caption

def create_plot(parameters, base_plot_params, data_map, title, data_key, func_prop, pt_marker=None):
    """Retrieve and prepare data to invoke plotting function
    """
    # Create specific plot parameters
    plot_params = dict(base_plot_params)

    # Retrieve plot data
    data_key_lower = data_key.lower()
    try:
        plot_params["data"] = data_map[data_key]
    except:
        try:
            plot_params["data"] = data_map[data_key_lower]
        except:
            print("*  WARNING: no data found for {}".format(data_key))
            return None

    # Retrieve variable names depending on function properties
    var_x_name = ''
    var_y_name = ''
    if isinstance(func_prop, str):
        var_x_name = data_key
        var_y_name = func_prop
    else:
        for p in func_prop:
            # Look for requested function name
            if p[0] in (data_key, data_key_lower):
                # Retrieve abscissa and/or ordinates when defined
                xy_names = p[1]
                if xy_names[0]:
                    var_x_name = xy_names[0]
                if xy_names[1]:
                    var_y_name = xy_names[1]
                break

    # Set other plot parameters
    plot_params["title"] = title
    plot_params["label"] = data_key
    plot_params["var_x_name"] = var_x_name
    plot_params["var_y_name"] = var_y_name
    plot_params["marker"] = pt_marker

    # Execute artifact generator and return its results
    return execute_request(parameters, plot_params)



# HEADER
#                     arg/Aggregation/argExodusAggregator.py
#               Automatic Report Generator (ARG) v. 1.0
#
# Copyright 2020 National Technology & Engineering Solutions of Sandia, LLC
# (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the U.S.
# Government retains certain rights in this software.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice,
#   this list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from this
#   software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# Questions? Visit gitlab.com/AutomaticReportGenerator/arg
#
# HEADER

import os
import math
import vtk
import vtkmodules.vtkFiltersExtraction as vtkFiltersExtraction

from arg.Common import argMath, argTools
from arg.Common.argMultiFontStringHelper import argMultiFontStringHelper
from arg.DataInterface.argDataInterface import argDataInterface
from arg.Generation import argVTK, argPlot
from arg.Aggregation.argAggregatorBase import argAggregatorBase


class argExodusAggregator(argAggregatorBase):
    """A class to aggregate information in a Exodus-specific way
    """

    def __init__(self, b, r):

        # Call superclass init
        super().__init__(b, r)


    def add_ignored_block_keys(self):
        """Add list of names of ignored blocks, when any
        """

        # Retrieve list of ignored block keys and bail out early if empty
        ignored_block_keys = self.RequestParameters.get("ignore_blocks")
        if not ignored_block_keys:
            return

        # Insert paragraph on ignored blocks
        ignored_string = argMultiFontStringHelper(self.Backend)
        ignored_string.append("N.B.: ", "bold")
        ignored_string.append("Blocks with ranks or names in ", "default")
        ignored_string.append("{}".format(", ".join([str(x) for x in ignored_block_keys])), "typewriter")
        ignored_string.append(" are ignored in this section.", "default")
        self.Backend.add_paragraph({"string": ignored_string})


    def compute_VTK_mesh_characteristics(self, vtk_data, block_id_to_flat):
        """Compute per-block topological and quality values of VTK mesh
        """

        # Initialize empty dicts for per-block storage of results
        n_verts, n_elems, t_elems, q_stats, q_histo = (
            {} for _ in range(5))

        # Build reverse lookup from flat to block indices
        block_flat_to_id = dict(
            (v, k) for k, vals in block_id_to_flat.items() for v in vals)

        # Iterate over non-empty blocks and extract those
        it = vtk_data.NewIterator()
        it.GoToFirstItem()
        while not it.IsDoneWithTraversal():
            # Retrieve flat index of current non-empty leaf
            idx = it.GetCurrentFlatIndex()

            # Skip ignored blocks
            if idx in block_flat_to_id:
                # Retrieve corresponding block ID and image file name
                b_id = block_flat_to_id[idx]

                # Extract mesh block
                extract = vtkFiltersExtraction.vtkExtractBlock()
                extract.SetInputData(vtk_data)
                extract.AddIndex(idx)
                extract.Update()
                mesh_block = extract.GetOutput().GetBlock(0)

                # Retrieve mesh block information and append it to global meta
                n_verts[b_id] = n_verts.get(b_id, 0) + mesh_block.GetNumberOfPoints()
                n_e_cur = n_elems.get(b_id, 0)
                n_e_blk = mesh_block.GetNumberOfCells()
                n_elems[b_id] = n_e_cur + n_e_blk
                elem_type_str, elem_q_type = argVTK.get_element_types(mesh_block)
                if True or elem_type_str:
                    t_elems.setdefault(b_id, elem_type_str)

                # Append mesh block quality when relevant
                if elem_q_type:
                    # Compute all desired quality statistics
                    for q_name, q_vtk in argVTK.quality_functions.items():
                        # Compute current block quality statistics and histogram
                        q_s_blk, q_h_blk = argVTK.get_mesh_quality(
                            mesh_block, q_vtk, elem_q_type, True)

                        # Update or create quality statistics for current block
                        q_stats[b_id] = argTools.update_or_create_dict_in_dict(
                            q_stats, b_id, q_name, q_s_blk,
                            argMath.aggregate_descriptive_statistics)

                        # Update or create quality histogram for current block
                        q_histo[b_id] = argTools.update_or_create_dict_in_dict(
                            q_histo, b_id, q_name, q_h_blk,
                            argMath.aggregate_histograms)

            # Iterate to next non-empty leaf
            it.GoToNextItem()

        # Return computed values
        return n_verts, n_elems, t_elems, q_stats, q_histo


    def add_block_quality_table(self, b_id, block_name, q_stats):
        """Add per-block mesh quality information when available
        """

        # Bail out early if no quality is available for block element type
        elt_quality_stats = q_stats.get(b_id)
        if not elt_quality_stats:
            # Mesh element type not supported by Verdict VTK
            print("*  WARNING: element type of block {} not supported by Verdict mesh quality library".format(
                b_id))
            return

        # Start new block quality statistics page
        self.Backend.add_page_break()
        quality_string = argMultiFontStringHelper(self.Backend)
        quality_string.append("Block ", "default")
        quality_string.append("{}".format(b_id), "typewriter")
        quality_string.append(" (", "default")
        quality_string.append(block_name, "typewriter")
        quality_string.append(") element quality", "default")
        self.Backend.add_subtitle({"title": quality_string})

        # Generate mesh quality table head
        head = [argMultiFontStringHelper(self.Backend) for _ in range(6)]
        head[0].append("Q", "calligraphic")
        head[1].append("min(", "default")
        head[1].append("Q", "calligraphic")
        head[1].append(')', "default")
        head[2].append("mu", "greek")
        head[2].append('(', "default")
        head[2].append("Q", "calligraphic")
        head[2].append(')', "default")
        head[3].append("max(", "default")
        head[3].append("Q", "calligraphic")
        head[3].append(')', "default")
        head[4].append("sigma", "greek")
        head[4].append('(', "default")
        head[4].append("Q", "calligraphic")
        head[4].append(')', "default")
        head[5].append("sigma", "greek")
        head[5].append('/', "default")
        head[5].append("mu", "greek")
        head[5].append('(', "default")
        head[5].append("Q", "calligraphic")
        head[5].append(')', "default")

        # Create mesh quality table body
        body = []
        for q_n, q_s in elt_quality_stats.items():
            # Bail out if no quality statistics were provided
            if not q_s:
                continue

            # Show standard deviation instead of variance
            stdev = math.sqrt(q_s[3]) if q_s[3] > 0. else 0.
            q_s[3] = stdev

            # Compute coefficient of variation
            q_s[4] = stdev / q_s[1] if q_s[1] > 1e-8 else float("inf")

            # Create new row for current quality name
            body.append([q_n] + ["{:.4g}".format(q) for q in q_s])


        # Assemble caption string
        stat_string = argMultiFontStringHelper(self.Backend)
        stat_string.append("Element quality statistics of block ", "default")
        stat_string.append(block_name, "typewriter")
        stat_string.append('.', "default")

        # Add table to report
        self.Backend.add_table(
            head,
            body,
            stat_string)


    def add_block_histograms(self, b_id, block_name, q_stats, q_histo, cv_t=1e-3):
        """Create and add histograms for given block
        """

        # Retrieve map of block histogram data
        histo_map = q_histo.get(b_id, {})
        for q_n, q_h in histo_map.items():
            # Bail out if no quality histogram was provided
            if not q_h:
                continue

            # Decide whether to show histogram depending on CoV
            cv_q = q_stats.get(b_id).get(q_n)[4]
            histo_string = argMultiFontStringHelper(self.Backend)
            histo_string.append("Histogram of {} element quality in block ".format(
                q_n), "default")
            histo_string.append(block_name, "typewriter")
            if cv_q < cv_t:
                # Quality spread is too tight for histogram
                histo_string.append(" is too narrow to be inserted (coefficient of variation: {} < {}).".format(
                    cv_q, cv_t), "default")
                self.Backend.add_paragraph({"string": histo_string})
            else:
                # Generate and insert histogram plot
                histo_name = argPlot.create_plot(
                    self.Backend.Parameters,
                    {"type": "histogram", "xyratio": 3.5},
                    histo_map,
                    "block {}".format(b_id),
                    q_n,
                    'element count',
                    'o')
                if histo_name:
                    histo_string.append('.', "default")
                    self.Backend.add_figure({
                        "figure_file": "{}.png".format(histo_name[0]),
                        "caption_string": histo_string,
                        "width": self.RequestParameters.get("histogram_width", "12cm")})


    def show_all_blocks(self, data, file_names):
        """Add surface rendering figures to the document for each mesh block
           for a specified point or cell data, scalar or vector variable
        """

        # Retrieve mandatory figure parameters
        try:
            model_data = data["model"]
            model_file = file_names["model"]
        except:
            print("*  WARNING: ignoring request: not enough parameters to show all blocks of input model")
            return

        # Get handle on input data in VTK form
        vtk_data = model_data.get_VTK_reader_output_data(0)
        if not vtk_data:
            print("*  WARNING: ignoring request: could not read VTK data from input model", model_file)
            return

        # Retrieve or create block names and IDs
        meta_data = model_data.get_meta_information()[0]
        block_names = meta_data.get("block names", [])
        block_IDs = meta_data.get("block IDs", [])

        # Retrieve comments if provided
        comments = argTools.map_composite_keys(
            self.RequestParameters, "string", self.Backend.Parameters.KeySeparator)

        # Generate block visualizations as needed
        variable = argVTK.argVTKAttribute(
            model_data, self.RequestParameters.get("time_step", -1))
        block_id_to_flat, block_images_and_captions = argVTK.all_blocks(
            self.Backend.Parameters,
            self.RequestParameters,
            model_data,
            variable,
            model_file)

        # Report about ignored blocks if any
        self.add_ignored_block_keys()

        # Compute topological and quality values of mesh
        n_verts, n_elems, t_elems, q_stats, q_histo = self.compute_VTK_mesh_characteristics(
            vtk_data, block_id_to_flat)

        # Create per-block pages
        for b_id, (base_name, caption) in sorted(block_images_and_captions.items()):
            # Retrieve block ID and name
            block_name = block_names[block_IDs.index(b_id)]

            # Start new block summary page
            self.Backend.add_page_break()
            block_string = argMultiFontStringHelper(self.Backend)
            block_string.append("Block ", "default")
            block_string.append("{}".format(b_id), "typewriter")
            block_string.append(" (", "default")
            block_string.append(block_name, "typewriter")
            block_string.append(") summary", "default")
            self.Backend.add_subtitle({"title": block_string})

            # Initialize storage for block properties table body
            body = [
                ["number of nodes", "{}".format(n_verts[b_id])],
                ["number of elements", "{}".format(n_elems[b_id])]]

            # Add block images only when element type is known
            if b_id in t_elems:
                # Add block images and element type to table body
                type_string = argMultiFontStringHelper(self.Backend)
                type_string.append(t_elems.get(b_id), "typewriter")
                body.append([
                    "type of first element in block",
                    type_string])
                self.Backend.add_figure({
                    "figure_file": base_name + ".png",
                    "caption_string": caption,
                    "width": self.RequestParameters.get("width", "12cm")})
            else:
                # Block element type is unknown
                body.append(["type of first element in block", "unknown"])

            # Generate table of block properties
            caption_string = argMultiFontStringHelper(self.Backend)
            caption_string.append("Properties of block ", "default")
            caption_string.append(block_name, "typewriter")
            caption_string.append('.', "default")
            self.Backend.add_table(
                ["property", "value"],
                body,
                caption_string,
                False,  # Do not do verbatim
                "hb!")  # Put table at bottom of page

            # Append per-block mesh quality information when available
            self.add_block_quality_table(b_id, block_name, q_stats)
            
            # Create and add quality histograms for current block
            self.add_block_histograms(b_id, block_name, q_stats, q_histo)

            # Append comment when defined for current block
            if not self.Backend.add_comment(comments, block_name):
                self.Backend.add_comment(comments, b_id)

        # Clear page after last block
        self.Backend.add_page_break()


    def show_all_modes(self, data, file_name):
        """Add surface rendering figures for all mode shapes
           for the Disp variable,
           with a normalized displacement with given factor,
           laid out in n_cols columns by n_rows rows
        """

        # Bail out early if no modes are present
        times = data.get_available_times()
        if not times:
            return

        # Get handle on input data in VTK form
        vtk_data = data.get_VTK_reader_output_data(0)
        if not vtk_data:
            print("*  WARNING: could not read VTK data from input model {}".format(
                file_name))
            return

        # Retrieve first sought mode
        times = data.get_available_times()

        # Determine extreme modes
        mode_min = self.RequestParameters.get("mode_min")
        if mode_min:
            mode_min = float(mode_min)
            for i_min, m in enumerate(times):
                if m > mode_min:
                    break
        else:
            i_min = 0
        mode_max = self.RequestParameters.get("mode_max")
        if mode_max:
            mode_max = float(mode_max)
            for i_max, m in enumerate(reversed(times)):
                if m < mode_max:
                    break
            i_max = len(times) - 1 - i_max
        else:
            i_max = len(times) - 1

        # Determine numbers of complete images and of remaining modes
        n_modes_per_image = self.RequestParameters.get("n_cols", 1) * self.RequestParameters.get("n_rows", 1)
        n_images = i_max - i_min + 1
        n_figs, n_rem = divmod(n_images, n_modes_per_image)

        # Generate images and captions for each for each block of modes
        all_base_names, all_captions = [], []
        variable = argVTK.argVTKAttribute(data, )
        mode_range = [i_min, i_min + n_modes_per_image - 1]

        # Start with all complete images
        for _ in range(n_figs):
            self.RequestParameters["range"] = mode_range
            base_name, caption = argVTK.many_modes(
                self.Backend.Parameters,
                self.RequestParameters,
                data,
                variable,
                file_name)
            all_base_names.append(base_name)
            all_captions.append(caption)
            mode_range[0] += n_modes_per_image
            mode_range[1] += n_modes_per_image

        # Create incomplete image if modes remain
        if n_rem:
            mode_range[1] = mode_range[0] + n_rem - 1
            self.RequestParameters["range"] = mode_range
            base_name, caption = argVTK.many_modes(
                self.Backend.Parameters,
                self.RequestParameters,
                data,
                variable,
                file_name)
            all_base_names.append(base_name)
            all_captions.append(caption)

        # Add corresponding modal frequencies table
        modes = []
        for i, m in enumerate(times):
            # Retain only those modes that are in range
            if i < i_min:
                continue
            if i > i_max:
                break

            # Append mode to table body list
            mode_string = argMultiFontStringHelper(self.Backend)
            mode_string.append("{:.1f}".format(m), "typewriter")
            modes.append(["{}".format(i), mode_string])

        # Assemble frequency string
        freq_string = argMultiFontStringHelper(self.Backend)
        freq_string.append("Modal frequencies", "default")
        if mode_min:
            freq_string.append(" greater than {:.1f}Hz".format(mode_min), "default")
        if mode_min and mode_max:
            freq_string.append(" and", "default")
        if mode_max:
            freq_string.append(" less than {:.1f}Hz".format(mode_max), "default")
        freq_string.append('.', "default")
        self.Backend.add_table(
            ["mode number", "frequency (Hz)"],
            modes,
            freq_string)

        # Create figures
        fig_width = self.RequestParameters.get("width")
        for base_name, caption_string in zip(all_base_names, all_captions):
            self.Backend.add_figure({
                "figure_file": base_name + ".png",
                "caption_string": caption_string,
                "width": fig_width})


    def show_enumerated_fields(self, data, file_name):
        """Add surface rendering figures to the document for a
           specified set of point or cell data, scalar or vector variables
        """

        # Set default figure width
        self.RequestParameters.setdefault("width", "12cm")

        # Create per-field pages
        for d in data:
            # Generate page header
            var_desc = "element" if d.AttributeBinding == "cell" else "node"
            var_desc += "-based {}".format(d.AttributeType)

            # Iterate over all available time
            for i, t in enumerate(d.get_available_times()):
                # Start with a new page for each time-step
                self.Backend.add_page_break()

                # Create variable page title
                variable_string = argMultiFontStringHelper(self.Backend)
                variable_string.append("Variable ", "default")
                variable_string.append(d.AttributeName, "typewriter")
                variable_string.append(" (" + var_desc, "default")

                # Append time when more than one time-step is availanle
                if len(d.get_available_times()) > 1:
                    variable_string.append(" at time ", "default")
                    variable_string.append("{:.6g}".format(t), "typewriter")

                # Create un-numbered subsection
                variable_string.append(')', "default")
                self.Backend.add_subtitle({"title": variable_string})

                # Get handle on input data in VTK form
                vtk_data = d.get_VTK_reader_output_data(i)
                if not vtk_data:
                    print("* {} WARNING: could not read VTK data from input model".format(file_name))
                    return

                # Generate clipped mesh surface view
                self.RequestParameters["time_step"] = i
                self.show_mesh_surface(d, file_name, True)

        # Clear page after last field
        self.Backend.add_page_break()


    def aggregate(self):
        """Decide which aggregation operation is to be performed
        """

        # Switch between different aggregation types
        try:
            request_name = self.RequestParameters["name"]
        except:
            print("*  WARNING: ignoring request: no aggregation methodn name")
        print("[argExodusAggregator] Processing {} request".format(request_name))

        # Operation show_all_blocks: one 4-view figure and one table per mesh block
        if request_name.startswith("show_all_blocks"):
            # Decide whether mesh edges are to be shown or not
            self.RequestParameters["edges"] = request_name.endswith("with_edges")

            # Add specific request parameters
            self.RequestParameters.setdefault("view_direction", ())

            # Retrieve model file
            model_file = self.RequestParameters.get("model")
            file_names = {"model": model_file}

            # Try to retrieve model data
            if model_file:
                model_data = argDataInterface.factory(
                    "ExodusII",
                    os.path.join(self.Backend.Parameters.DataDir, model_file),
                    self.RequestParameters.get("var_name", ''))
            else:
                model_data = None
            data = {"model": model_data}

            # Aggregate
            if model_data:
                self.show_all_blocks(data, file_names)

        # Operation show_enumerated_fields: one figure page per field
        elif request_name == "show_enumerated_fields":
            # Decide whether mesh edges are to be shown or not
            self.RequestParameters["edges"] = request_name.endswith("with_edges")

            # Add specific request parameters
            self.RequestParameters.setdefault("view_direction", ())
            var_names = self.RequestParameters.get("var_names")
            if not var_names:
                print("*  WARNING: no variable names provided for {}".format(
                    request_name))
                return

            # Get handle on data
            file_name = self.RequestParameters["model"]
            data = [argDataInterface.factory(
                "ExodusII",
                os.path.join(self.Backend.Parameters.DataDir, file_name), v, False)
                for v in var_names]

            # Aggregate
            if data:
                self.show_enumerated_fields(data, file_name)

        # Operation show_all_modes: one figure every n_cols x n_rows modes
        elif request_name.startswith("show_all_modes"):
            # Decide whether mesh edges are to be shown or not
            self.RequestParameters["edges"] = request_name.endswith("with_edges")

            # Add specific request parameters
            self.RequestParameters.setdefault("view_direction", ())
            self.RequestParameters.setdefault("displacement", 2.)
            self.RequestParameters.setdefault("n_cols", 4)
            self.RequestParameters.setdefault("n_rows", 4)

            # Get handle on data
            file_name = self.RequestParameters["model"]
            data = argDataInterface.factory(
                "ExodusII",
                os.path.join(self.Backend.Parameters.DataDir, file_name),
                self.RequestParameters.get("var_name", "Disp"),
                False)

            # Aggregate
            self.show_all_modes(data, file_name)

        # Operation show_mesh_surface: create one figure for the entire mesh
        elif request_name.startswith("show_mesh_surface"):
            # Decide whether mesh edges are to be shown or not
            self.RequestParameters["edges"] = request_name.endswith("with_edges")

            # Add specific request parameters
            self.RequestParameters.setdefault("view_direction", ())

            # Get handle on data
            file_name = self.RequestParameters["model"]
            data = argDataInterface.factory(
                "ExodusII",
                os.path.join(self.Backend.Parameters.DataDir, file_name),
                self.RequestParameters.get("var_name", ''))

            # Aggregate
            if data:
                self.show_mesh_surface(data, file_name)
